__all__ = ("apply_zp_norm", "PlotHandler", "BasePlotter")

import os
import warnings

import matplotlib.pyplot as plt
import numpy as np

import rubin_sim.maf.utils as utils


def apply_zp_norm(metric_value, plot_dict):
    if "zp" in plot_dict:
        if plot_dict["zp"] is not None:
            metric_value = metric_value - plot_dict["zp"]
    if "norm_val" in plot_dict:
        if plot_dict["norm_val"] is not None:
            metric_value = metric_value / plot_dict["norm_val"]
    return metric_value


class BasePlotter:
    """
    Serve as the base type for MAF plotters and example of API.
    """

    def __init__(self):
        self.plot_type = None
        # This should be included in every subsequent default_plot_dict
        # (assumed to be present).
        self.default_plot_dict = {
            "title": None,
            "xlabel": None,
            "label": None,
            "labelsize": None,
            "fontsize": None,
            "figsize": None,
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """
        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer.
        user_plot_dict: `dict`
            Dictionary of plot parameters set by user
            (overrides default values).
        fig : `matplotlib.figure.Figure`
            Matplotlib figure number to use. Default = None, starts new figure.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
           Figure with the plot.
        """
        pass


class PlotHandler:
    """Create plots from a single or series of metric bundles.

    Parameters
    ----------
    out_dir : `str`, optional
        Directory to save output plots.
    results_db : `rubin_sim.maf.ResultsDb`, optional
        ResultsDb into which to record plot location and information.
    savefig : `bool`, optional
        Flag for saving images to disk (versus create and return
        to caller).
    fig_format : `str`, optional
        Figure format to use to save full-size output. Default PDF.
    dpi : `int`, optional
        DPI to save output figures to disk at. (for matplotlib figures).
    thumbnail : `bool`, optional
        Flag for saving thumbnails (reduced size pngs) to disk.
    trim_whitespace : `bool`, optional
        Flag for trimming whitespace option to matplotlib output figures.
        Default True, usually doesn't need to be changed.
    """

    def __init__(
        self,
        out_dir=".",
        results_db=None,
        savefig=True,
        fig_format="pdf",
        dpi=600,
        thumbnail=True,
        trim_whitespace=True,
    ):
        self.out_dir = out_dir
        self.results_db = results_db
        self.savefig = savefig
        self.fig_format = fig_format
        self.dpi = dpi
        self.trim_whitespace = trim_whitespace
        self.thumbnail = thumbnail
        self.filtercolors = {
            "u": "cyan",
            "g": "g",
            "r": "y",
            "i": "r",
            "z": "m",
            "y": "k",
            " ": None,
        }
        self.filterorder = {" ": -1, "u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}

    def set_metric_bundles(self, m_bundles):
        """
        Set the metric bundle or bundles (list or dictionary).
        Reuse the PlotHandler by resetting this reference.
        The metric bundles have to have the same slicer.
        """
        self.m_bundles = []
        # Try to add the metricBundles in filter order.
        if isinstance(m_bundles, dict):
            for m_b in m_bundles.values():
                vals = m_b.file_root.split("_")
                forder = [self.filterorder.get(f, None) for f in vals if len(f) == 1]
                forder = [o for o in forder if o is not None]
                if len(forder) == 0:
                    forder = len(self.m_bundles)
                else:
                    forder = forder[-1]
                self.m_bundles.insert(forder, m_b)
            self.slicer = self.m_bundles[0].slicer
        else:
            for m_b in m_bundles:
                vals = m_b.file_root.split("_")
                forder = [self.filterorder.get(f, None) for f in vals if len(f) == 1]
                forder = [o for o in forder if o is not None]
                if len(forder) == 0:
                    forder = len(self.m_bundles)
                else:
                    forder = forder[-1]
                self.m_bundles.insert(forder, m_b)
            self.slicer = self.m_bundles[0].slicer
        for m_b in self.m_bundles:
            if m_b.slicer.slicer_name != self.slicer.slicer_name:
                raise ValueError("MetricBundle items must have the same type of slicer")
        self._combine_metric_names()
        self._combine_run_names()
        self._combine_metadata()
        self._combine_constraints()
        self.set_plot_dicts(reset=True)

    def set_plot_dicts(self, plot_dicts=None, plot_func=None, reset=False):
        """
        Set or update the plot_dict for the (possibly joint) plots.

        Resolution is: (from lowest to higher)
        auto-generated items (colors/labels/titles)
        < anything previously set in the plot_handler
        < defaults set by the plotter
        < explicitly set items in the metricBundle plot_dict
        < explicitly set items in the plot_dicts list passed to this method.
        """
        if reset:
            # Have to explicitly set each dictionary to a (separate)
            # blank dictionary.
            self.plot_dicts = [{} for b in self.m_bundles]

        if isinstance(plot_dicts, dict):
            # We were passed a single dictionary, not a list.
            plot_dicts = [plot_dicts] * len(self.m_bundles)

        auto_label_list = self._build_legend_labels()
        auto_color_list = self._build_colors()
        auto_cbar = self._build_cbar_format()
        auto_title = self._build_title()
        if plot_func is not None:
            auto_xlabel, auto_ylabel = self._build_x_ylabels(plot_func)

        # Loop through each bundle and generate a plot_dict for it.
        for i, bundle in enumerate(self.m_bundles):
            # First use the auto-generated values.
            tmp_plot_dict = {}
            tmp_plot_dict["title"] = auto_title
            tmp_plot_dict["label"] = auto_label_list[i]
            tmp_plot_dict["color"] = auto_color_list[i]
            tmp_plot_dict["cbar_format"] = auto_cbar
            # Update that with anything previously set in the plot_handler.
            tmp_plot_dict.update(self.plot_dicts[i])
            # Then override with plot_dict items set explicitly
            # based on the plot type.
            if plot_func is not None:
                tmp_plot_dict["xlabel"] = auto_xlabel
                tmp_plot_dict["ylabel"] = auto_ylabel
                # Replace auto-generated plot dict items with things
                #  set by the plotter_defaults, if they are not None.
                plotter_defaults = plot_func.default_plot_dict
                for k, v in plotter_defaults.items():
                    if v is not None:
                        tmp_plot_dict[k] = v
            # Then add/override based on the bundle plot_dict parameters
            # if they are set.
            tmp_plot_dict.update(bundle.plot_dict)
            # Finally, override with anything set explicitly by the user.
            if plot_dicts is not None:
                tmp_plot_dict.update(plot_dicts[i])
            # And save this new dictionary back in the class.
            self.plot_dicts[i] = tmp_plot_dict

        # Check that the plot_dicts do not conflict.
        self._check_plot_dicts()

    def _combine_metric_names(self):
        """
        Combine metric names.
        """
        # Find the unique metric names.
        self.metric_names = set()
        for m_b in self.m_bundles:
            self.metric_names.add(m_b.metric.name)
        # Find a pleasing combination of the metric names.
        order = ["u", "g", "r", "i", "z", "y"]
        if len(self.metric_names) == 1:
            joint_name = " ".join(self.metric_names)
        else:
            # Split each unique name into a list to see
            # if we can merge the names.
            name_lengths = [len(x.split()) for x in self.metric_names]
            name_lists = [x.split() for x in self.metric_names]
            # If the metric names are all the same length, see
            # if we can combine any parts.
            if len(set(name_lengths)) == 1:
                joint_name = []
                for i in range(name_lengths[0]):
                    tmp = set([x[i] for x in name_lists])
                    # Try to catch special case of filters and
                    # put them in order.
                    if tmp.intersection(order) == tmp:
                        filterlist = ""
                        for f in order:
                            if f in tmp:
                                filterlist += f
                        joint_name.append(filterlist)
                    else:
                        # Otherwise, just join and put into joint_name.
                        joint_name.append("".join(tmp))
                joint_name = " ".join(joint_name)
            # If the metric names are not the same length,
            # just join everything.
            else:
                joint_name = " ".join(self.metric_names)
        self.joint_metric_names = joint_name

    def _combine_run_names(self):
        """
        Combine runNames.
        """
        self.run_names = set()
        for m_b in self.m_bundles:
            self.run_names.add(m_b.run_name)
        self.joint_run_names = " ".join(self.run_names)

    def _combine_metadata(self):
        """
        Combine info_label.
        """
        info_label = set()
        for m_b in self.m_bundles:
            info_label.add(m_b.info_label)
        self.info_label = info_label
        # Find a pleasing combination of the info_label.
        if len(info_label) == 1:
            self.joint_metadata = " ".join(info_label)
        else:
            order = ["u", "g", "r", "i", "z", "y"]
            # See if there are any subcomponents we can combine,
            # splitting on some values we expect to separate
            # info_label clauses.
            splitmetas = []
            for m in self.info_label:
                # Try to split info_label into separate phrases
                # (filter / proposal / constraint..).
                if " and " in m:
                    m = m.split(" and ")
                elif ", " in m:
                    m = m.split(", ")
                else:
                    m = [
                        m,
                    ]
                # Strip white spaces from individual elements.
                m = set([im.strip() for im in m])
                splitmetas.append(m)
            # Look for common elements and
            # separate from the general info_label.
            common = set.intersection(*splitmetas)
            diff = [x.difference(common) for x in splitmetas]
            # Now look within the 'diff' elements and
            # see if there are any common words to split off.
            diffsplit = []
            for d in diff:
                if len(d) > 0:
                    m = set([x.split() for x in d][0])
                else:
                    m = set()
                diffsplit.append(m)
            diffcommon = set.intersection(*diffsplit)
            diffdiff = [x.difference(diffcommon) for x in diffsplit]
            # If the length of any of the 'differences' is 0,
            # then we should stop and not try to subdivide.
            lengths = [len(x) for x in diffdiff]
            if min(lengths) == 0:
                # Sort them in order of length
                # (so it goes 'g', 'g dithered', etc.)
                tmp = []
                for d in diff:
                    tmp.append(list(d)[0])
                diff = tmp
                xlengths = [len(x) for x in diff]
                idx = np.argsort(xlengths)
                diffdiff = [diff[i] for i in idx]
                diffcommon = []
            else:
                # diffdiff is the part where we might expect
                # our filter values to appear;
                # try to put this in order.
                diffdiff_ordered = []
                diffdiff_end = []
                for f in order:
                    for d in diffdiff:
                        if len(d) == 1:
                            if list(d)[0] == f:
                                diffdiff_ordered.append(d)
                for d in diffdiff:
                    if d not in diffdiff_ordered:
                        diffdiff_end.append(d)
                diffdiff = diffdiff_ordered + diffdiff_end
                diffdiff = [" ".join(c) for c in diffdiff]
            # And put it all back together.
            combo = (
                ", ".join(["".join(c) for c in diffdiff])
                + " "
                + " ".join(["".join(d) for d in diffcommon])
                + " "
                + " ".join(["".join(e) for e in common])
            )
            self.joint_metadata = combo

    def _combine_constraints(self):
        """
        Combine the constraints.
        """
        constraints = set()
        for m_b in self.m_bundles:
            if m_b.constraint is not None:
                constraints.add(m_b.constraint)
        self.constraints = "; ".join(constraints)

    def _build_title(self):
        """
        Build a plot title from the metric names, runNames and info_label.
        """
        # Create a plot title from the unique parts of
        # the metric/run_name/info_label.
        plot_title = ""
        if len(self.run_names) == 1:
            plot_title += list(self.run_names)[0]
        if len(self.info_label) == 1:
            plot_title += " " + list(self.info_label)[0]
        if len(self.metric_names) == 1:
            plot_title += ": " + list(self.metric_names)[0]
        if plot_title == "":
            # If there were more than one of everything above,
            # use joint info_label and metricNames.
            plot_title = self.joint_metadata + " " + self.joint_metric_names
        return plot_title

    def _build_x_ylabels(self, plot_func, len_max=25):
        """
        Build a plot x and y label.

        Parameters
        ----------
        len_max : `int`, optional
            If the xlabel starts longer than this, add the units as a newline.
        """
        if plot_func.plot_type == "BinnedData":
            if len(self.m_bundles) == 1:
                m_b = self.m_bundles[0]
                if len(m_b.slicer.slice_col_name) < len_max:
                    xlabel = m_b.slicer.slice_col_name + " (" + m_b.slicer.slice_col_units + ")"
                else:
                    xlabel = m_b.slicer.slice_col_name + " \n(" + m_b.slicer.slice_col_units + ")"
                ylabel = m_b.metric.name + " (" + m_b.metric.units + ")"
            else:
                xlabel = set()
                for m_b in self.m_bundles:
                    xlabel.add(m_b.slicer.slice_col_name)
                xlabel = ", ".join(xlabel)
                ylabel = self.joint_metric_names
        elif plot_func.plot_type == "MetricVsH":
            if len(self.m_bundles) == 1:
                m_b = self.m_bundles[0]
                ylabel = m_b.metric.name + " (" + m_b.metric.units + ")"
            else:
                ylabel = self.joint_metric_names
            xlabel = "H (mag)"
        else:
            if len(self.m_bundles) == 1:
                m_b = self.m_bundles[0]
                xlabel = m_b.metric.name
                if m_b.metric.units is not None:
                    if len(m_b.metric.units) > 0:
                        if len(xlabel) < len_max:
                            xlabel += " (" + m_b.metric.units + ")"
                        else:
                            xlabel += "\n(" + m_b.metric.units + ")"
                ylabel = None
            else:
                xlabel = self.joint_metric_names
                ylabel = set()
                for m_b in self.m_bundles:
                    if "ylabel" in m_b.plot_dict:
                        ylabel.add(m_b.plot_dict["ylabel"])
                if len(ylabel) == 1:
                    ylabel = list(ylabel)[0]
                else:
                    ylabel = None
        return xlabel, ylabel

    def _build_legend_labels(self):
        """
        Build a set of legend labels,
        using the parts of the run_name/info_label/metricNames that change.
        """
        if len(self.m_bundles) == 1:
            return [None]
        labels = []
        for m_b in self.m_bundles:
            if "label" in m_b.plot_dict:
                label = m_b.plot_dict["label"]
            else:
                label = ""
                if len(self.run_names) > 1:
                    label += m_b.run_name
                if len(self.info_label) > 1:
                    label += " " + m_b.info_label
                if len(self.metric_names) > 1:
                    label += " " + m_b.metric.name
            labels.append(label)
        return labels

    def _build_colors(self):
        """
        Try to set an appropriate range of colors for the metric Bundles.
        """
        if len(self.m_bundles) == 1:
            if "color" in self.m_bundles[0].plot_dict:
                return [self.m_bundles[0].plot_dict["color"]]
            else:
                return ["b"]
        colors = []
        for m_b in self.m_bundles:
            color = "b"
            if "color" in m_b.plot_dict:
                color = m_b.plot_dict["color"]
            else:
                if m_b.constraint is not None:
                    # If the filter is part of the sql constraint, we'll
                    #  try to use that first.
                    if "filter" in m_b.constraint:
                        vals = m_b.constraint.split('"')
                        for v in vals:
                            if len(v) == 1:
                                # Guess that this is the filter value
                                if v in self.filtercolors:
                                    color = self.filtercolors[v]
            colors.append(color)
        # If we happened to end up with the same color throughout
        #  (say, the metrics were all in the same filter)
        #  then go ahead and generate random colors.
        if (len(self.m_bundles) > 1) and (len(np.unique(colors)) == 1):
            colors = [
                np.random.rand(
                    3,
                )
                for m_b in self.m_bundles
            ]
        return colors

    def _build_cbar_format(self):
        """
        Set the color bar format.
        """
        cbar_format = None
        if len(self.m_bundles) == 1:
            if self.m_bundles[0].metric.metric_dtype == "int":
                cbar_format = "%d"
        else:
            metric_dtypes = set()
            for m_b in self.m_bundles:
                metric_dtypes.add(m_b.metric.metric_dtype)
            if len(metric_dtypes) == 1:
                if list(metric_dtypes)[0] == "int":
                    cbar_format = "%d"
        return cbar_format

    def _build_file_root(self, outfile_suffix=None):
        """
        Build a root filename for plot outputs.
        If there is only one metricBundle,
        this is equal to the metricBundle fileRoot + outfile_suffix.
        For multiple metricBundles,
        this is created from the runNames, info_label and metric names.

        If you do not wish to use the automatic filenames,
        then you could set 'savefig' to False and
        save the file manually to disk,
        using the plot figure numbers returned by 'plot'.
        """
        if len(self.m_bundles) == 1:
            outfile = self.m_bundles[0].file_root
        else:
            outfile = "_".join([self.joint_run_names, self.joint_metric_names, self.joint_metadata])
            outfile += "_" + self.m_bundles[0].slicer.slicer_name[:4].upper()
        if outfile_suffix is not None:
            outfile += "_" + outfile_suffix
        outfile = utils.name_sanitize(outfile)
        return outfile

    def _build_display_dict(self):
        """
        Generate a display dictionary.
        This is most useful for when there are many metricBundles
        being combined into a single plot.
        """
        if len(self.m_bundles) == 1:
            return self.m_bundles[0].display_dict
        else:
            display_dict = {}
            group = set()
            subgroup = set()
            order = 0
            for m_b in self.m_bundles:
                group.add(m_b.display_dict["group"])
                subgroup.add(m_b.display_dict["subgroup"])
                if order < m_b.display_dict["order"]:
                    order = m_b.display_dict["order"] + 1
            display_dict["order"] = order
            if len(group) > 1:
                display_dict["group"] = "Comparisons"
            else:
                display_dict["group"] = list(group)[0]
            if len(subgroup) > 1:
                display_dict["subgroup"] = "Comparisons"
            else:
                display_dict["subgroup"] = list(subgroup)[0]

            display_dict["caption"] = (
                "%s metric(s) calculated on a %s grid, for opsim runs %s, for info_label values of %s."
                % (
                    self.joint_metric_names,
                    self.m_bundles[0].slicer.slicer_name,
                    self.joint_run_names,
                    self.joint_metadata,
                )
            )

            return display_dict

    def _check_plot_dicts(self):
        """
        Check to make sure there are no conflicts in the plot_dicts
        that are being used in the same subplot.
        """
        # Check that the length is OK
        if len(self.plot_dicts) != len(self.m_bundles):
            raise ValueError(
                "plot_dicts (%i) must be same length as mBundles (%i)"
                % (len(self.plot_dicts), len(self.m_bundles))
            )

        # These are the keys that need to match (or be None)
        keys2_check = ["xlim", "ylim", "color_min", "color_max", "title"]

        # Identify how many subplots there are.
        # If there are more than one, just don't change anything.
        # This assumes that if there are more than one,
        # the plot_dicts are actually all compatible.
        subplots = set()
        for pd in self.plot_dicts:
            if "subplot" in pd:
                subplots.add(pd["subplot"])

        # Now check subplots are consistent.
        if len(subplots) <= 1:
            reset_keys = []
            for key in keys2_check:
                values = [pd[key] for pd in self.plot_dicts if key in pd]
                if len(np.unique(values)) > 1:
                    # We will reset some of the keys to the default,
                    # but for some we should do better.
                    if key.endswith("Max"):
                        for pd in self.plot_dicts:
                            pd[key] = np.max(values)
                    elif key.endswith("Min"):
                        for pd in self.plot_dicts:
                            pd[key] = np.min(values)
                    elif key == "title":
                        title = self._build_title()
                        for pd in self.plot_dicts:
                            pd["title"] = title
                    else:
                        warnings.warn(
                            'Found more than one value to be set for "%s" in the plot_dicts.' % (key)
                            + " Will reset to default value. (found values %s)" % values
                        )
                        reset_keys.append(key)
            # Reset the most of the keys to defaults;
            # this can generally be done safely.
            for key in reset_keys:
                for pd in self.plot_dicts:
                    pd[key] = None

    def plot(
        self,
        plot_func,
        plot_dicts=None,
        display_dict=None,
        outfile_root=None,
        outfile_suffix=None,
    ):
        """
        Create a plot for the active metric bundles (self.set_metric_bundles).

        Parameters
        ----------
        plot_func : `rubin_sim.plots.BasePlotter`
            The plotter to use to make the figure.
        plot_dicts : `list` of [`dict`], optional
            List of plot_dicts for each metric bundle.
            Can use these to override individual metric bundle colors, etc.
        display_dict : `dict`, optional
            Information to save to resultsDb to accompany the figure on the
            show_maf pages. Generally set automatically. Includes a caption.
        outfile_root : `str`, optional
            Output filename. Generally set automatically, but can be
            overriden (such as when output filenames get too long).
        outfile_suffix : `str`, optional
            A suffix to add to the end of the default output filename.
            Useful when creating a series of plots, such as for a movie.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            The plot.
        """
        if not plot_func.object_plotter:
            # Check that metric_values type and plotter are compatible
            # (most are float/float, but some plotters expect object data ..
            # and some only do sometimes).
            for m_b in self.m_bundles:
                if m_b.metric.metric_dtype == "object":
                    metric_is_color = m_b.plot_dict.get("metric_is_color", False)
                    if not metric_is_color:
                        warnings.warn("Cannot plot object metric values with this plotter.")
                        return

        # Update x/y labels using plot_type.
        self.set_plot_dicts(plot_dicts=plot_dicts, plot_func=plot_func, reset=False)
        # Set outfile name.
        if outfile_root is None:
            outfile = self._build_file_root(outfile_suffix)
        else:
            outfile = outfile_root
        plot_type = plot_func.plot_type
        if len(self.m_bundles) > 1:
            plot_type = "Combo" + plot_type
        # Make plot.
        fig = None
        for m_b, plot_dict in zip(self.m_bundles, self.plot_dicts):
            if m_b.metric_values is None:
                # Skip this metricBundle.
                msg = 'MetricBundle (%s) has no "metric_values".' % (m_b.file_root)
                msg += " Either the values have not been calculated or they have been deleted."
                warnings.warn(msg)
            elif np.size(np.where(~m_b.metric_values.mask)) == 0:
                msg = "MetricBundle (%s) has no unmasked metric_values, skipping plots." % (m_b.file_root)
                warnings.warn(msg)
            else:
                fig = plot_func(m_b.metric_values, m_b.slicer, plot_dict, fig=fig)
        # Add a legend if more than one metricValue is being plotted
        # or if legendloc is specified.
        legend_loc = None
        if "legend_loc" in self.plot_dicts[0]:
            legend_loc = self.plot_dicts[0]["legend_loc"]
        if len(self.m_bundles) > 1:
            try:
                legend_loc = self.plot_dicts[0]["legend_loc"]
            except KeyError:
                legend_loc = "upper right"
        if legend_loc is not None:
            # Activate expected figure and write legend.
            plt.figure(fig)
            plt.legend(loc=legend_loc, fancybox=True, fontsize="smaller")
        # Add the super title if provided.
        if "suptitle" in self.plot_dicts[0]:
            plt.suptitle(self.plot_dicts[0]["suptitle"])
        # Save to disk and file info to results_db if desired.
        if self.savefig:
            if display_dict is None:
                display_dict = self._build_display_dict()
            self.save_fig(
                fig,
                outfile,
                plot_type,
                self.joint_metric_names,
                self.slicer.slicer_name,
                self.joint_run_names,
                self.constraints,
                self.joint_metadata,
                display_dict,
            )
        return fig

    def save_fig(
        self,
        fig,
        outfile_root,
        plot_type,
        metric_name,
        slicer_name,
        run_name,
        constraint,
        info_label,
        display_dict=None,
    ):
        plot_file = outfile_root + "_" + plot_type + "." + self.fig_format
        if fig is None:
            warnings.warn(f"Trying to save figure to {plot_file} but" "figure is None. Skipping.")
            return
        if self.trim_whitespace:
            fig.savefig(
                os.path.join(self.out_dir, plot_file),
                dpi=self.dpi,
                bbox_inches="tight",
                format=self.fig_format,
            )
        else:
            fig.savefig(
                os.path.join(self.out_dir, plot_file),
                dpi=self.dpi,
                format=self.fig_format,
            )
        # Generate a png thumbnail.
        if self.thumbnail:
            thumb_file = "thumb." + outfile_root + "_" + plot_type + ".png"
            fig.savefig(os.path.join(self.out_dir, thumb_file), dpi=72, bbox_inches="tight")
        # Save information about the file to results_db.
        if self.results_db:
            if display_dict is None:
                display_dict = {}
            metric_id = self.results_db.update_metric(
                metric_name, slicer_name, run_name, constraint, info_label, None
            )
            self.results_db.update_display(metric_id=metric_id, display_dict=display_dict, overwrite=False)
            self.results_db.update_plot(metric_id=metric_id, plot_type=plot_type, plot_file=plot_file)



__all__ = ("StarCountMetric",)

import numpy as np

from rubin_sim.maf.metrics import BaseMetric

from .star_counts import starcount

# Example for CountMetric
# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Motivation: The distances to stars in LSST will be significant enough
# that the structure of the Galaxy will be readily apparent because of
# its influence on the number of stars in a given field.
# Any metric concerned with the number of potential objects to be
# detected will need to feature not only the effects of the cadence
# but also the number of objects per field.
# This metric identifies the number of stars in a given field in a
# particular distance range.
# D1 and D2 are the close and far distances in parsecs.
# Requires StarCounts.StarCounts

# NOTE
# There are stellar luminosity function maps available within MAF
# that may supersede these StarCount functions


class StarCountMetric(BaseMetric):
    """Find the number of stars in a given field between d1 and d2 in parsecs.

    This metric uses the stellar distance and luminosity equations
    contributed by Mike Lund, which are based on the Galfast model.
    There are some imposed limitations on the expected magnitudes
    of the stars included for the metric, based on assuming saturation
    at 16th magnitude and not considering stars with magnitude
    uncertainties greater than 0.03 (based on photometry/m5 alone).


    Parameters
    ----------
    d1 : `float`
        d1 in parsecs
    d2 : `float`
        d2 in parsecs
    """

    def __init__(self, d1=100, d2=1000, **kwargs):
        self.d1 = d1
        self.d2 = d2
        super(StarCountMetric, self).__init__(col=[], **kwargs)

    def run(self, data_slice, slice_point=None):
        self.dec_col = np.degrees(data_slice[0][3])
        self.ra_col = np.degrees(data_slice[0][2])
        return starcount(self.ra_col, self.dec_col, self.d1, self.d2)



__all__ = ("YearCoverageMetric",)

import numpy as np

from .base_metric import BaseMetric


class YearCoverageMetric(BaseMetric):
    """Count the number of `bins` covered by night_col.

    The default `bins` cover years 0 to 10.
    Handy for checking that a point on the sky gets observed every year,
    as the default settings result in the metric returning the number years
    in the data_slice (when used with a HealpixSlicer).

    Parameters
    ----------
    night_col : `str`, opt
        Data column to histogram. Default 'night'.
    bins : `np.ndarray`, (N,), opt
        Bins to use in the histogram. Default corresponds to years 0-10
        (with 365.25 nights per year).
    units : `str`, opt
        Units to use for the metric result. Default 'N years'.

    Returns
    -------
    nbins : `int`
        Number of histogram bins where the histogram value is greater than 0.
        Typically this will be the number of years in the 'night_col'.
    """

    def __init__(self, night_col="night", bins=None, units=None, **kwargs):
        self.night_col = night_col
        if bins is None:
            self.bins = np.arange(0, np.ceil(365.25 * 10.0), 365.25) - 0.5
        else:
            self.bins = bins

        if units is None:
            units = "N years"

        super().__init__([night_col], units=units)

    def run(self, data_slice, slice_point):
        hist, be = np.histogram(data_slice[self.night_col], bins=self.bins)
        result = np.where(hist > 0)[0].size
        return result



# Example for TripletMetric
# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 9/6/2014
# Motivation: The detection of nonperiodic transient events can be thought of
# as most simply being accomplished by a set of three observations,
# one before the event occurs, a second after the event has begun,
# and a third to confirm the event is real.
# This metric identifies the number of triplets that will occur.
# DelMin and DelMax set the smallest and largest intervals that can occur
# between the first and second point and between the second and third point.
# This can be set to reflect the timescales for various events.
# RatioMax and RatioMin set constraints on how similar the two intervals
# must be. RatioMin can never be less than 1.

__all__ = ("TripletMetric", "TripletBandMetric")

import numpy as np

from rubin_sim.maf.metrics import BaseMetric


class TripletMetric(BaseMetric):
    """Find the number of 'triplets' of three images taken in any band,
    based on user-selected minimum and maximum intervals (in hours),
    as well as constraining the ratio of the two exposures intervals.
    Triplets are not required to be consecutive observations and
    may be overlapping.
    """

    def __init__(self, time_col="expMJD", **kwargs):
        self.time_col = time_col
        self.delmin = kwargs.pop("DelMin", 1) / 24.0  # convert minutes to hours
        self.delmax = kwargs.pop("DelMax", 12) / 24.0  # convert minutes to hours
        self.ratiomax = kwargs.pop("RatioMax", 1000)
        self.ratiomin = kwargs.pop("RatioMin", 1)
        super(TripletMetric, self).__init__(col=[self.time_col], **kwargs)

    def run(self, data_slice, slice_point=None):
        times = data_slice[self.time_col]
        times = times - 49378  # change times to smaller numbers
        delmax = self.delmax
        delmin = self.delmin
        total = 0
        # iterate over every exposure time
        for counter, time in enumerate(times):
            # calculate the window to look for all possible second points in
            minmax = [time + delmin, time + delmax]
            index2 = np.where((minmax[0] < times) & (times < minmax[1]))[0]
            # iterate over every middle exposure
            for middleindex in index2:
                timeb = times[middleindex]
                # calculate the window to look for all possible third points
                minmax2 = [timeb + delmin, timeb + delmax]
                index3 = np.where((times > minmax2[0]) & (times < minmax2[1]))[0]
                newadd = np.size(index3)
                total = total + newadd  # add all triplets with same
                # first two observations to total
        return total


class TripletBandMetric(BaseMetric):
    """Find the number of 'triplets' of three images taken in the same band,
    based on user-selected minimum and maximum intervals (in hours),
    as well as constraining the ratio of the two exposures intervals.
    Triplets are not required to be consecutive observations and may be
    overlapping.
    """

    def __init__(self, time_col="expMJD", filter_col="filter", **kwargs):
        self.time_col = time_col
        self.filter_col = filter_col
        self.delmin = kwargs.pop("DelMin", 1) / 24.0
        self.delmax = kwargs.pop("DelMax", 12) / 24.0
        self.ratiomax = kwargs.pop("RatioMax", 1000)
        self.ratiomin = kwargs.pop("RatioMin", 1)
        super(TripletBandMetric, self).__init__(col=[self.time_col, self.filter_col], **kwargs)
        self.reduce_order = {
            "Bandu": 0,
            "Bandg": 1,
            "Bandr": 2,
            "Bandi": 3,
            "Bandz": 4,
            "Bandy": 5,
        }

    def run(self, data_slice, slice_point=None):
        times = data_slice[self.time_col]
        times = times - 49378  # change times to smaller numbers
        bands = data_slice[self.filter_col]
        bandset = ["u", "g", "r", "i", "z", "y"]  # list of possible bands
        timedict = {}
        delmax = self.delmax
        delmin = self.delmin
        ratiomax = self.ratiomax
        ratiomin = self.ratiomin
        bandcounter = {
            "u": 0,
            "g": 0,
            "r": 0,
            "i": 0,
            "z": 0,
            "y": 0,
        }  # define zeroed out counter
        # iterate over each bandpass
        for band in bandset:
            indexlist = np.where(band == bands)
            timedict[band] = times[indexlist]
            # create a data set of all exposures for a single band
            timeband = timedict[band]
            # iterate over every exposure time
            for counter, time in enumerate(timeband):
                # calculate the window to look for all possible second points
                minmax = [time + delmin, time + delmax]
                index2 = np.where((minmax[0] < timeband) & (timeband < minmax[1]))[0]
                # iterate over every middle exposure
                for middleindex in index2:
                    timeb = timeband[middleindex]
                    # calculate the window to look for all
                    # possible third points
                    minmax2 = [timeb + delmin, timeb + delmax]
                    index3 = np.where((timeband > minmax2[0]) & (timeband < minmax2[1]))[0]
                    # iterate over last exposure of triplet
                    for lastindex in index3:
                        timec = timeband[lastindex]
                        # calculate intervals for T1 to T2 and T2 to T3,
                        # and take ratio
                        delt1 = timeb - time
                        delt2 = timec - timeb
                        ratio = np.max([delt1, delt2]) / np.min([delt1, delt2])
                        # check if ratio is within restrictions
                        # (ratio should never be < 1 )
                        if ratiomin < ratio < ratiomax:
                            bandcounter[band] = bandcounter[band] + 1
        return bandcounter

    def reduce_bandall(self, bandcounter):
        return np.sum(list(bandcounter.values()))

    def reduce_bandu(self, bandcounter):
        return bandcounter["u"]

    def reduce_bandg(self, bandcounter):
        return bandcounter["g"]

    def reduce_bandr(self, bandcounter):
        return bandcounter["r"]

    def reduce_bandi(self, bandcounter):
        return bandcounter["i"]

    def reduce_bandz(self, bandcounter):
        return bandcounter["z"]

    def reduce_bandy(self, bandcounter):
        return bandcounter["y"]



__all__ = ("StringCountMetric",)

from collections import Counter

import numpy as np

from .base_metric import BaseMetric


class Keylookerupper:
    """Helper object to unpack dictionary values as reduceFunction results."""

    def __init__(self, key="blank", name=None):
        self.key = key
        self.__name__ = name

    def __call__(self, indict):
        return np.max(indict[self.key])


class StringCountMetric(BaseMetric):
    """Count up the number of times each string appears in a column.

    Dynamically builds reduce functions for each unique string value,
    so summary stats can be named the same as strings in the
    simData array without knowing the values of those strings ahead of time.


    Parameters
    ----------
    metric_name : `str`, opt
        Name of the metric.
    col : `str`, opt
        Column name that has strings to look at.
    percent : `bool`, opt
        Normalize and return results as percents rather than raw count.
    clip_end : `bool`
        Clip if the end of a string if it ends with a comma and number.
    """

    def __init__(
        self, metric_name="stringCountMetric", col="filter", percent=False, clip_end=False, **kwargs
    ):
        if percent:
            units = "percent"
        else:
            units = "count"
        self.percent = percent
        cols = [col]
        super().__init__(cols, metric_name, units=units, metric_dtype=object, **kwargs)
        self.col = col
        self.clip_end = clip_end

    def run(self, data_slice, slice_point=None):

        # If we need to clip off trailing integer
        if self.clip_end:
            replace_col = []
            for val in data_slice[self.col]:
                if ", " in val:
                    chunks = val.split(", ")
                    if chunks[-1].isdigit():
                        new_val = ", ".join(chunks[0:-1])
                        replace_col.append(new_val)
                    else:
                        replace_col.append(val)
                else:
                    replace_col.append(val)
            data_slice[self.col] = replace_col

        counter = Counter(data_slice[self.col])
        # convert to a numpy array
        lables = list(counter.keys())
        # Numpy can't handle empty string as a dtype
        lables = [x if x != "" else "blank" for x in lables]
        metric_value = np.zeros(1, dtype=list(zip(lables, [float] * len(counter.keys()))))
        for key in counter:
            if key == "":
                metric_value["blank"] = counter[key]
            else:
                metric_value[key] = counter[key]
        if self.percent:
            norm = sum(metric_value[0]) / 100.0
            # Not sure I really like having to loop here,
            # but the dtype is inflexible
            for key in metric_value.dtype.names:
                metric_value[key] = metric_value[key] / norm

        # Now to dynamically set up the reduce functions
        for i, key in enumerate(metric_value.dtype.names):
            name = key
            self.reduce_funcs[name] = Keylookerupper(key=key, name=name)
            self.reduce_order[name] = i

        return metric_value



import unittest

import numpy as np

import rubin_sim.maf.utils as utils


class StellarMagsTest(unittest.TestCase):
    """
    Test the example stellar colors code
    """

    def test_sm(self):
        keys = [
            "O",
            "B",
            "A",
            "F",
            "G",
            "K",
            "M",
            "HeWD_25200_80",
            "WD_11000_85",
            "WD_3000_85",
        ]
        filter_names = ["u", "g", "r", "i", "z", "y"]

        # Check each type returns the correct format
        for key in keys:
            result = utils.stellar_mags(key)
            for fn in filter_names:
                self.assertIn(fn, result)
                self.assertTrue(
                    (isinstance(result[fn], float)) | (isinstance(result[fn], np.float64)),
                    msg="result is neither a float nor a numpy float64",
                )

        # Check the exception gets raised
        self.assertRaises(ValueError, utils.stellar_mags, "ack")

        # Check the mags get fainter
        for st in keys:
            mags = utils.stellar_mags(st)
            mags2 = utils.stellar_mags(st, rmag=20.0)
        for key in mags:
            self.assertLess(mags[key], mags2[key])


if __name__ == "__main__":
    unittest.main()



__all__ = ("HourglassPlot",)

import matplotlib.pyplot as plt
import numpy as np

from .plot_handler import BasePlotter


class HourglassPlot(BasePlotter):
    def __init__(self):
        self.plot_type = "Hourglass"
        self.object_plotter = True
        self.default_plot_dict = {
            "title": None,
            "xlabel": "Night - min(Night)",
            "ylabel": "Hours from local midnight",
            "figsize": None,
        }
        self.filter2color = {
            "u": "purple",
            "g": "blue",
            "r": "green",
            "i": "cyan",
            "z": "orange",
            "y": "red",
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """
        Generate the hourglass plot
        """
        if slicer.slicer_name != "HourglassSlicer":
            raise ValueError("HourglassPlot is for use with hourglass slicers")

        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        ax = fig.add_subplot(111)
        pernight = metric_value[0]["pernight"]
        perfilter = metric_value[0]["perfilter"]

        y = (perfilter["mjd"] - perfilter["midnight"]) * 24.0
        dmin = np.floor(np.min(perfilter["mjd"]))
        for i in np.arange(0, perfilter.size, 2):
            ax.plot(
                [perfilter["mjd"][i] - dmin, perfilter["mjd"][i + 1] - dmin],
                [y[i], y[i + 1]],
                self.filter2color[perfilter["filter"][i]],
            )
        for i, key in enumerate(["u", "g", "r", "i", "z", "y"]):
            ax.text(
                1.05,
                0.9 - i * 0.07,
                key,
                color=self.filter2color[key],
                transform=ax.transAxes,
            )
        # ax.plot(pernight['mjd'] - dmin,
        # (pernight['twi6_rise'] - pernight['midnight']) * 24.,
        #        'blue', label=r'6$^\circ$ twilight')
        # ax.plot(pernight['mjd'] - dmin,
        # (pernight['twi6_set'] - pernight['midnight']) * 24.,
        #        'blue')
        ax.plot(
            pernight["mjd"] - dmin,
            (pernight["twi12_rise"] - pernight["midnight"]) * 24.0,
            "yellow",
            label=r"12$^\circ$ twilight",
        )
        ax.plot(
            pernight["mjd"] - dmin,
            (pernight["twi12_set"] - pernight["midnight"]) * 24.0,
            "yellow",
        )
        ax.plot(
            pernight["mjd"] - dmin,
            (pernight["twi18_rise"] - pernight["midnight"]) * 24.0,
            "red",
            label=r"18$^\circ$ twilight",
        )
        ax.plot(
            pernight["mjd"] - dmin,
            (pernight["twi18_set"] - pernight["midnight"]) * 24.0,
            "red",
        )
        ax.plot(
            pernight["mjd"] - dmin,
            pernight["moonPer"] / 100.0 - 7,
            "black",
            label="Moon",
        )
        ax.set_xlabel(plot_dict["xlabel"])
        ax.set_ylabel(plot_dict["ylabel"])
        ax.set_title(plot_dict["title"])

        # draw things in with lines if we are only plotting one night
        if len(pernight["mjd"]) == 1:
            ax.axhline(
                (pernight["twi6_rise"] - pernight["midnight"]) * 24.0,
                color="blue",
                label=r"6$^\circ$ twilight",
            )
            ax.axhline((pernight["twi6_set"] - pernight["midnight"]) * 24.0, color="blue")
            ax.axhline(
                (pernight["twi12_rise"] - pernight["midnight"]) * 24.0,
                color="yellow",
                label=r"12$^\circ$ twilight",
            )
            ax.axhline((pernight["twi12_set"] - pernight["midnight"]) * 24.0, color="yellow")
            ax.axhline(
                (pernight["twi18_rise"] - pernight["midnight"]) * 24.0,
                color="red",
                label=r"18$^\circ$ twilight",
            )
            ax.axhline((pernight["twi18_set"] - pernight["midnight"]) * 24.0, color="red")

        return fig



__all__ = ("NightPointingMetric",)

import numpy as np
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation, get_body, get_sun
from astropy.time import Time
from rubin_scheduler.utils import Site

from .base_metric import BaseMetric


class NightPointingMetric(BaseMetric):
    """
    Gather relevant information for a night to plot.
    """

    def __init__(
        self,
        alt_col="altitude",
        az_col="azimuth",
        filter_col="filter",
        mjd_col="observationStartMJD",
        metric_name="NightPointing",
        telescope="LSST",
        **kwargs,
    ):
        cols = [alt_col, az_col, filter_col, mjd_col]
        super(NightPointingMetric, self).__init__(
            col=cols, metric_name=metric_name, metric_dtype="object", **kwargs
        )
        self.telescope = Site(name=telescope)
        self.alt_col = alt_col
        self.az_col = az_col
        self.filter_col = filter_col
        self.mjd_col = mjd_col

        self.location = EarthLocation(
            lat=self.telescope.latitude_rad * u.rad,
            lon=self.telescope.longitude_rad * u.rad,
            height=self.telescope.height * u.m,
        )

    def run(self, data_slice, slice_point=None):
        pad = 30.0 / 60.0 / 24.0
        mjd_min = data_slice[self.mjd_col].min() - pad
        mjd_max = data_slice[self.mjd_col].max() + pad

        # How often to plot the moon and things
        step = 20.0 / 60.0 / 24.0
        mjds = Time(np.arange(mjd_min, mjd_max + step, step), format="mjd")

        aa = AltAz(location=self.location, obstime=mjds)

        moon_coords = get_body("moon", mjds).transform_to(aa)
        sun_coords = get_sun(mjds).transform_to(aa)

        moon_alts = np.array(moon_coords.alt.rad)
        moon_azs = np.array(moon_coords.az.rad)
        mjds = np.array(mjds)
        sun_alts = np.array(sun_coords.alt.rad)
        sun_azs = np.array(sun_coords.az.rad)

        return {
            "data_slice": data_slice,
            "moon_alts": moon_alts,
            "moon_azs": moon_azs,
            "mjds": mjds,
            "sun_alts": sun_alts,
            "sun_azs": sun_azs,
        }



__all__ = ("MafTracking",)

import os
from collections import OrderedDict

import numpy as np

from rubin_sim.maf.utils import get_sim_data

from .maf_run_results import MafRunResults


class MafTracking:
    """Hold and serve the MAF tracking (sqlite) database content for
    the show_maf web server.

    Parameters
    ----------
    database : `str`, optional
        Path to the sqlite tracking database file.
        If None, looks for `trackingDb_sqlite.db` default file in the
        current directory.
    """

    def __init__(self, database=None):
        if database is None:
            database = os.path.join(os.getcwd(), "trackingDb_sqlite.db")
        self.tracking_db = database

        # Read in the tracking database.
        cols = [
            "maf_run_id",
            "run_name",
            "run_group",
            "maf_comment",
            "run_comment",
            "db_file",
            "maf_dir",
            "run_version",
            "run_date",
            "maf_version",
            "maf_date",
        ]
        self.runs = get_sim_data(self.tracking_db, "", cols, table_name="runs")
        self.runs = self.sort_runs(self.runs, order=["maf_run_id", "run_name", "maf_comment"])
        self.runs_page = {}

    def run_info(self, run):
        """Get the tracking database information relevant for a given run
        in a format that the jinja2 templates for show_maf  can use.

        Parameters
        ----------
        run : `np.ndarray`, (1,)
           One line from self.runs

        Returns
        -------
        run_info : `OrderedDict`
            Ordered dict version of the numpy structured array.
        """
        runInfo = OrderedDict()
        maf_dir = os.path.relpath(run["maf_dir"], start=os.path.dirname(self.tracking_db))
        runInfo["Run Name"] = run["run_name"]
        runInfo["Group"] = run["run_group"]
        runInfo["Maf Comment"] = run["maf_comment"]
        runInfo["Run Comment"] = run["run_comment"]
        runInfo["RunDb File"] = [
            os.path.relpath(run["db_file"]),
            os.path.split(run["db_file"])[1],
        ]
        runInfo["ResultsDb"] = os.path.join(maf_dir, "resultsDb_sqlite.db")
        runInfo["maf_dir"] = maf_dir
        runInfo["sched_version"] = run["run_version"]
        runInfo["sched_date"] = run["run_date"]
        runInfo["maf_version"] = run["maf_version"]
        runInfo["maf_date"] = run["maf_date"]
        return runInfo

    def sort_runs(self, runs, order=["run_name", "maf_comment", "maf_run_id"]):
        """Sort  the numpy array of run data.

        Parameters
        ----------
        runs : `np.ndarray`, (N,)
           The runs from self.runs to sort.
        order : `list` [`str`]
           The fields to use to sort the runs array.

        Returns
        -------
        runs : `numpy.NDarray`
           A sorted numpy array.
        """
        return np.sort(runs, order=order)

    def get_run(self, maf_run_id):
        """Set up a mafRunResults object to read and handle the data
        from a single individual run.
        Caches the mafRunResults object, meaning the metric information from
        a particular run is only read once from disk.

        Parameters
        ----------
        maf_run_id : `int`
           maf_run_id value in the tracking database
           corresponding to a particular MAF run.

        Returns
        -------
        runPage : `MafRunResults`
           A MafRunResults object containing the information for this run.
           Stored internally in self.runs_page dict, but also passed
           back to the tornado server.
        """
        if not isinstance(maf_run_id, int):
            if isinstance(maf_run_id, dict):
                maf_run_id = int(maf_run_id["maf_run_id"][0][0])
            if isinstance(maf_run_id, list):
                maf_run_id = int(maf_run_id[0])
        if maf_run_id in self.runs_page:
            return self.runs_page[maf_run_id]
        match = self.runs["maf_run_id"] == maf_run_id
        maf_dir = self.runs[match]["maf_dir"][0]
        run_name = self.runs[match]["run_name"][0]
        if run_name == "NULL":
            run_name = None
        self.runs_page[maf_run_id] = MafRunResults(maf_dir, run_name)
        return self.runs_page[maf_run_id]



__all__ = ("find_telescopes", "NFollowStacker")


import numpy as np

from .base_stacker import BaseStacker
from .coord_stackers import ra_dec2_alt_az


def find_telescopes(min_size=3.0):
    """Finds telescopes larger than min_size,
    from list of large telescopes based on
    http://astro.nineplanets.org/bigeyes.html.

    Returns
    -------
    np.recarray
        Array of large telescopes with columns [aperture, name, lat, lon].
    """
    # Aperture  Name Location http://astro.nineplanets.org/bigeyes.html
    telescopes = [
        [10.4, "Gran Canarias", "La Palma"],
        [10.0, "Keck", "Mauna Kea"],
        [10.0, "Keck II", "Mauna Kea"],
        [9.2, "SALT", "South African Astronomical Observatory"],
        [9.2, "Hobby-Eberly", "Mt. Fowlkes"],
        [8.4, "Large Binocular Telescope", "Mt. Graham"],
        [8.3, "Subaru", "Mauna Kea"],
        [8.2, "Antu", "Cerro Paranal"],
        [8.2, "Kueyen", "Cerro Paranal"],
        [8.2, "Melipal", "Cerro Paranal"],
        [8.2, "Yepun", "Cerro Paranal"],
        [8.1, "Gemini North", "Mauna Kea"],
        [8.1, "Gemini South", "Cerro Pachon"],
        [6.5, "MMT", "Mt. Hopkins"],
        [6.5, "Walter Baade", "La Serena"],
        [6.5, "Landon Clay", "La Serena"],
        [6.0, "Bolshoi Teleskop Azimutalnyi", "Nizhny Arkhyz"],
        [6.0, "LZT", "British Columbia"],
        [5.0, "Hale", "Palomar Mountain"],
        [4.3, "Dicovery Channel", "Lowell Observatory"],
        [4.2, "William Herschel", "La Palma"],
        [4.2, "SOAR", "Cerro Pachon"],
        [4.2, "LAMOST", "Xinglong Station"],
        [4.0, "Victor Blanco", "Cerro Tololo"],
        [4.0, "Vista", "Cerro Paranal"],
        [3.9, "Anglo-Australian", "Coonabarabran"],
        [3.8, "Mayall", "Kitt Peak"],
        [3.8, "UKIRT", "Mauna Kea"],
        [3.6, "360", "Cerro La Silla"],
        [3.6, "Canada-France-Hawaii", "Mauna Kea"],
        [3.6, "Telescopio Nazionale Galileo", "La Palma"],
        [3.5, "MPI-CAHA", "Calar Alto"],
        [3.5, "New Technology", "Cerro La Silla"],
        [3.5, "ARC", "Apache Point"],
        [3.5, "WIYN", "Kitt Peak"],
        [3.0, "Shane", "Mount Hamilton"],
        [3.0, "NASA IRTF", "Mauna Kea"],
    ]

    scopes = np.zeros(
        len(telescopes),
        dtype=list(zip(["aperture", "name", "lat", "lon"], [float, (np.str_, 38), float, float])),
    )

    # name, lat (S negative), lon (W negative)
    observatories = [
        ["Cerro Paranal", -24, 38, -70, 24],
        ["Nizhny Arkhyz", 43, 39, 41, 26],
        ["Cerro La Silla", -29, 15, -70, 44],
        ["Lowell Observatory", 35, 12, -111, 40],
        ["Apache Point", 32, 47, -105, 49],
        ["Mount Hamilton", 37, 21, -121, 38],
        ["South African Astronomical Observatory", -32, 23, 20, 49],
        ["Cerro Pachon", -30, 20, -70, 59],
        ["Coonabarabran", -31, 17, 149, 0o4],
        ["Mt. Fowlkes", 30, 40, -104, 1],
        ["La Palma", 28, 46, -17, 53],
        ["Mt. Graham", 32, 42, -109, 53],
        ["Calar Alto", 37, 13, -2, 33],
        ["British Columbia", 49, 17, -122, 34],
        ["Kitt Peak", 31, 57, -111, 37],
        ["La Serena", -30, 10, -70, 48],
        ["Palomar Mountain", 33, 21, -116, 52],
        ["Xinglong Station", 40, 23, 105, 50],
        ["Mt. Hopkins", 31, 41, -110, 53],
        ["Cerro Tololo", -30, 10, -70, 49],
        ["Mauna Kea", 19, 50, -155, 28],
    ]

    # Make a nice little dict to look up the observatory positions
    obs = {}
    for i, ob in enumerate(observatories):
        obs[ob[0]] = [
            (np.abs(ob[1]) + ob[2] / 60.0) * (ob[1] / np.abs(ob[1])),
            (np.abs(ob[3]) + ob[4] / 60.0) * (ob[3] / np.abs(ob[3])),
        ]

    for i, telescope in enumerate(telescopes):
        scopes["aperture"][i] = telescope[0]
        scopes["name"][i] = telescope[1]
        scopes["lat"][i], scopes["lon"][i] = obs[telescope[2]]

    scopes = scopes[np.where(scopes["aperture"] >= min_size)]
    return scopes


class NFollowStacker(BaseStacker):
    """Add the number of telescopes ('nObservatories') that could
    follow up any visit at (any of the) times in timeStep,
    specifying the minimum telescope size (in meters) and airmass limit.

    Parameters
    ----------
    minSize: float, optional
        The minimum telescope aperture to use, in meters. Default 3.0.
    airmass_limit: float, optional
        The maximum airmass allowable at the follow-up observatory.
        Default 2.5.
    time_steps: np.array or list of floats, optional
        The timesteps to check for followup opportunities, in hours.
        Default is np.arange(0.5, 12., 3.0).
    mjd_col: str, optional
        The exposure MJD column name. Default 'observationStartMJD'.
    ra_col: str, optional
        The RA column name. Default 'fieldRA'.
    dec_col: str, optional
        The Dec column name. Default 'fieldDec'.
    raDecDeg: bool, optional
        Flag whether RA/Dec are in degrees (True) or radians (False).
    """

    cols_added = ["nObservatories"]

    def __init__(
        self,
        min_size=3.0,
        airmass_limit=2.5,
        time_steps=np.arange(0.5, 12.0, 3.0),
        mjd_col="observationStartMJD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
    ):
        self.mjd_col = mjd_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.degrees = degrees
        self.cols_added_dtypes = [int]
        self.cols_req = [self.mjd_col, self.ra_col, self.dec_col]
        self.units = ["#"]
        self.airmass_limit = airmass_limit
        self.time_steps = time_steps
        self.telescopes = find_telescopes(min_size=min_size)

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        sim_data["nObservatories"] = 0
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        for obs in self.telescopes:
            obs_got_it = np.zeros(len(sim_data[self.ra_col]), int)
            obs_lon = np.radians(obs["lon"])
            obs_lat = np.radians(obs["lat"])
            for step in self.time_steps:
                alt, az = ra_dec2_alt_az(
                    ra,
                    dec,
                    obs_lon,
                    obs_lat,
                    sim_data[self.mjd_col] + step / 24.0,
                    altonly=True,
                )
                airmass = 1.0 / (np.cos(np.pi / 2.0 - alt))
                followed = np.where((airmass <= self.airmass_limit) & (airmass >= 1.0))
                # If the observatory got an observation, save this
                # into obs_got_it.
                # obs_got_it will be 1 if ANY of the times got an observation.
                obs_got_it[followed] = 1
            # If an observatory got an observation, count it in nObservatories.
            sim_data["nObservatories"] += obs_got_it
        return sim_data



__all__ = ("glanceBatch",)

import warnings

from rubin_scheduler.scheduler.utils import EuclidOverlapFootprint
from rubin_scheduler.utils import ddf_locations

import rubin_sim.maf.metric_bundles as metric_bundles
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as stackers

from .col_map_dict import col_map_dict
from .common import standard_summary
from .hourglass_batch import hourglassPlots
from .slew_batch import slewBasics


def glanceBatch(
    colmap=None,
    run_name="run_name",
    nside=64,
    filternames=("u", "g", "r", "i", "z", "y"),
    nyears=10,
    pairnside=32,
    sql_constraint=None,
    slicer_camera="LSST",
):
    """Generate a handy set of metrics that give a quick overview
    of how well a survey performed.
    This is a meta-set of other batches, to some extent.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    run_name : `str`, optional
        The name of the simulated survey.
    nside : `int`, optional
        The nside for the healpix slicers.
    filternames : `list` of `str`, optional
        The list of individual filters to use when running metrics.
        There is always an all-visits version of the metrics run as well.
    nyears : `int`, optional
        How many years to attempt to make hourglass plots for
    pairnside : `int`, optional
        nside to use for the pair fraction metric
        (it's slow, so nice to use lower resolution)
    sql_constraint : `str` or None, optional
        Additional SQL constraint to apply to all metrics.
    slicer_camera : `str`
        Sets which spatial slicer to use. options are 'LSST' and 'ComCam'

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if isinstance(colmap, str):
        raise ValueError("colmap must be a dictionary, not a string")

    if colmap is None:
        colmap = col_map_dict()

    bundle_list = []

    if sql_constraint is None:
        sqlC = ""
    else:
        sqlC = "(%s) and" % sql_constraint

    if slicer_camera == "LSST":
        spatial_slicer = slicers.HealpixSlicer
    elif slicer_camera == "ComCam":
        spatial_slicer = slicers.HealpixComCamSlicer
    else:
        raise ValueError("Camera must be LSST or Comcam")

    sql_per_filt = ['%s %s="%s"' % (sqlC, colmap["filter"], filtername) for filtername in filternames]
    sql_per_and_all_filters = [sql_constraint] + sql_per_filt

    standardStats = standard_summary()
    subsetPlots = [plots.HealpixSkyMap(), plots.HealpixHistogram()]

    # Super basic things
    displayDict = {"group": "Basic Stats", "order": 1}
    sql = sql_constraint
    slicer = slicers.UniSlicer()
    # Length of Survey
    metric = metrics.FullRangeMetric(col=colmap["mjd"], metric_name="Length of Survey (days)")
    bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
    bundle_list.append(bundle)

    # Total number of filter changes
    metric = metrics.NChangesMetric(col=colmap["filter"], order_by=colmap["mjd"])
    bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
    bundle_list.append(bundle)

    # Total open shutter fraction
    metric = metrics.OpenShutterFractionMetric(
        slew_time_col=colmap["slewtime"],
        exp_time_col=colmap["exptime"],
        visit_time_col=colmap["visittime"],
    )
    bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
    bundle_list.append(bundle)

    # Total effective exposure time
    metric = metrics.MeanMetric(col="t_eff")
    teff_stacker = stackers.TeffStacker(normed=True)
    for sql in sql_per_and_all_filters:
        bundle = metric_bundles.MetricBundle(
            metric, slicer, sql, stacker_list=[teff_stacker], display_dict=displayDict
        )
        bundle_list.append(bundle)

    # Number of observations, all and each filter
    metric = metrics.CountMetric(col=colmap["mjd"], metric_name="Number of Exposures")
    for sql in sql_per_and_all_filters:
        bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
        bundle_list.append(bundle)

    # The alt/az plots of all the pointings
    slicer = spatial_slicer(
        nside=nside,
        lat_col=colmap["alt"],
        lon_col=colmap["az"],
        lat_lon_deg=colmap["raDecDeg"],
        use_cache=False,
    )
    metric = metrics.CountMetric(colmap["mjd"], metric_name="Nvisits as function of Alt/Az")
    plotFuncs = [plots.LambertSkyMap()]

    plotDict = {"norm": "log"}
    for sql in sql_per_and_all_filters:
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            plot_funcs=plotFuncs,
            display_dict=displayDict,
            plot_dict=plotDict,
        )
        bundle_list.append(bundle)

    # alt az of long gaps
    sql = "scheduler_note = 'long'"
    metric = metrics.CountMetric(colmap["mjd"], metric_name="Nvisits long")
    bundle = metric_bundles.MetricBundle(
        metric,
        slicer,
        sql,
        plot_funcs=plotFuncs,
        display_dict=displayDict,
        plot_dict=plotDict,
    )
    bundle_list.append(bundle)

    sql = "scheduler_note like 'blob_long%'"
    metric = metrics.CountMetric(colmap["mjd"], metric_name="Nvisits blob long")
    bundle = metric_bundles.MetricBundle(
        metric,
        slicer,
        sql,
        plot_funcs=plotFuncs,
        display_dict=displayDict,
        plot_dict=plotDict,
    )
    bundle_list.append(bundle)

    sql = "scheduler_note like '%neo%' or scheduler_note like '%near_sun%'"
    metric = metrics.CountMetric(colmap["mjd"], metric_name="Nvisits twilight near sun")
    bundle = metric_bundles.MetricBundle(
        metric,
        slicer,
        sql,
        plot_funcs=plotFuncs,
        display_dict=displayDict,
        plot_dict=plotDict,
    )
    bundle_list.append(bundle)

    # alt,az pf ToO

    sql = "scheduler_note like 'ToO%'"
    metric = metrics.CountMetric(colmap["mjd"], metric_name="Nvisits long")
    bundle = metric_bundles.MetricBundle(
        metric,
        slicer,
        sql,
        plot_funcs=plotFuncs,
        display_dict=displayDict,
        plot_dict=plotDict,
    )
    bundle_list.append(bundle)

    # Things to check per night
    # Open Shutter per night
    displayDict = {"group": "Pointing Efficency", "order": 2}
    slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=1, bin_min=-0.5)
    metric = metrics.OpenShutterFractionMetric(
        slew_time_col=colmap["slewtime"],
        exp_time_col=colmap["exptime"],
        visit_time_col=colmap["visittime"],
    )
    sql = sql_constraint
    bundle = metric_bundles.MetricBundle(
        metric, slicer, sql, summary_metrics=standardStats, display_dict=displayDict
    )
    bundle_list.append(bundle)

    # Number of filter changes per night
    slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=1, bin_min=-0.5)
    metric = metrics.NChangesMetric(
        col=colmap["filter"], order_by=colmap["mjd"], metric_name="Filter Changes"
    )
    bundle = metric_bundles.MetricBundle(
        metric, slicer, sql, summary_metrics=standardStats, display_dict=displayDict
    )
    bundle_list.append(bundle)

    # A few basic maps
    # Number of observations, coadded depths
    extended_stats = standardStats.copy()
    extended_stats.append(metrics.AreaSummaryMetric(decreasing=True, metric_name="top18k"))
    extended_stats.append(metrics.PercentileMetric(col="metricdata", percentile=10))
    displayDict = {"group": "Basic Maps", "order": 3}
    slicer = spatial_slicer(
        nside=nside,
        lat_col=colmap["dec"],
        lon_col=colmap["ra"],
        lat_lon_deg=colmap["raDecDeg"],
    )
    metric = metrics.CountMetric(col=colmap["mjd"])
    plotDict = {"percentile_clip": 95.0}
    for sql in sql_per_and_all_filters:
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            summary_metrics=extended_stats,
            display_dict=displayDict,
            plot_dict=plotDict,
        )
        bundle_list.append(bundle)

    metric = metrics.Coaddm5Metric(m5_col=colmap["fiveSigmaDepth"])
    for sql in sql_per_and_all_filters:
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            summary_metrics=extended_stats,
            display_dict=displayDict,
        )
        bundle_list.append(bundle)

    # Let's look at two years
    displayDict = {"group": "Roll Check", "order": 1}
    rolling_metrics = []
    rolling_metrics.append(metrics.CountMetric(col=colmap["mjd"], metric_name="Year1.0Count"))
    rolling_metrics.append(metrics.CountMetric(col=colmap["mjd"], metric_name="Year2.5Count"))
    rolling_metrics.append(metrics.CountMetric(col=colmap["mjd"], metric_name="Year3.5Count"))
    rolling_sqls = []
    rolling_sqls.append("night < 365.25")
    rolling_sqls.append("night > %f and night < %f" % (365.25 * 2.5, 365.25 * 3.5))
    rolling_sqls.append("night > %f and night < %f" % (365.25 * 3.5, 365.25 * 4.5))
    for metric, sql in zip(rolling_metrics, rolling_sqls):
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            summary_metrics=extended_stats,
            plot_dict=plotDict,
            display_dict=displayDict,
        )
        bundle_list.append(bundle)

    # Make a cumulative plot of a WFD spot
    sql = "scheduler_note not like '%NEO%' and scheduler_note not like '%near_sun%'"
    uslicer = slicers.UserPointsSlicer(ra=0, dec=-20)
    metric = metrics.CumulativeMetric()
    metricb = metric_bundles.MetricBundle(
        metric,
        uslicer,
        sql,
        plot_funcs=[plots.XyPlotter()],
        run_name=run_name,
        display_dict=displayDict,
    )
    metricb.summary_metrics = []
    bundle_list.append(metricb)

    # Checking a few basic science things
    # Maybe check astrometry, observation pairs, SN
    plotDict = {"percentile_clip": 95.0}
    displayDict = {"group": "Science", "subgroup": "Astrometry", "order": 4}

    stackerList = []
    stacker = stackers.ParallaxFactorStacker(
        ra_col=colmap["ra"],
        dec_col=colmap["dec"],
        degrees=colmap["raDecDeg"],
        date_col=colmap["mjd"],
    )
    stackerList.append(stacker)

    astrom_stats = [
        metrics.AreaSummaryMetric(decreasing=False, metric_name="best18k"),
        metrics.PercentileMetric(col="metricdata", percentile=90),
    ]
    displayDict["caption"] = r"Parallax precision of an $r=20$ flat SED star"
    metric = metrics.ParallaxMetric(
        m5_col=colmap["fiveSigmaDepth"],
        filter_col=colmap["filter"],
        seeing_col=colmap["seeingGeom"],
    )
    sql = sql_constraint
    bundle = metric_bundles.MetricBundle(
        metric,
        slicer,
        sql,
        plot_funcs=subsetPlots,
        display_dict=displayDict,
        stacker_list=stackerList,
        plot_dict=plotDict,
        summary_metrics=astrom_stats,
    )
    bundle_list.append(bundle)
    displayDict["caption"] = r"Proper motion precision of an $r=20$ flat SED star"
    metric = metrics.ProperMotionMetric(
        m5_col=colmap["fiveSigmaDepth"],
        mjd_col=colmap["mjd"],
        filter_col=colmap["filter"],
        seeing_col=colmap["seeingGeom"],
    )
    bundle = metric_bundles.MetricBundle(
        metric,
        slicer,
        sql,
        plot_funcs=subsetPlots,
        display_dict=displayDict,
        plot_dict=plotDict,
        summary_metrics=astrom_stats,
    )
    bundle_list.append(bundle)

    # Solar system stuff
    displayDict["caption"] = "Fraction of observations that are in pairs"
    displayDict["subgroup"] = "Solar System"

    sql = '%s (filter="g" or filter="r" or filter="i")' % sqlC
    pairSlicer = slicers.HealpixSlicer(
        nside=pairnside,
        lat_col=colmap["dec"],
        lon_col=colmap["ra"],
        lat_lon_deg=colmap["raDecDeg"],
    )
    metric = metrics.PairFractionMetric(mjd_col=colmap["mjd"])
    bundle = metric_bundles.MetricBundle(
        metric, pairSlicer, sql, plot_funcs=subsetPlots, display_dict=displayDict
    )
    bundle_list.append(bundle)

    # stats from the scheduler_note column
    if "scheduler_note" in colmap.keys():
        displayDict = {"group": "Basic Stats", "subgroup": "Percent root stats"}
        metric = metrics.StringCountMetric(
            col=colmap["scheduler_note_root"], percent=True, metric_name="Percents", clip_end=False
        )
        sql = ""
        slicer = slicers.UniSlicer()
        bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
        bundle_list.append(bundle)
        displayDict["subgroup"] = "Count root Stats"
        metric = metrics.StringCountMetric(
            col=colmap["scheduler_note_root"], metric_name="Counts", clip_end=False
        )
        bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
        bundle_list.append(bundle)

    # For pairs and twilights
    if "scheduler_note" in colmap.keys():
        displayDict = {"group": "Basic Stats", "subgroup": "Percent stats"}
        metric = metrics.StringCountMetric(
            col=colmap["scheduler_note"], percent=True, metric_name="Percents", clip_end=False
        )
        sql = (
            "scheduler_note like 'pair%%' or scheduler_note like 'twilight%%' or scheduler_note like 'blob%%'"
        )
        slicer = slicers.UniSlicer()
        bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
        bundle_list.append(bundle)
        displayDict["subgroup"] = "Count Stats"
        metric = metrics.StringCountMetric(col=colmap["scheduler_note"], metric_name="Counts", clip_end=False)
        bundle = metric_bundles.MetricBundle(metric, slicer, sql, display_dict=displayDict)
        bundle_list.append(bundle)

    # DDF progress
    ddf_surveys = ddf_locations()
    displayDict["group"] = "DDF"
    displayDict["subgroup"] = ""
    for ddf in ddf_surveys:
        label = ddf.replace("DD:", "")
        sql = 'scheduler_note like "%s%%"' % ("DD:" + label)
        slicer = slicers.UniSlicer()
        metric = metrics.CumulativeMetric()
        metricb = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            plot_funcs=[plots.XyPlotter()],
            run_name=run_name,
            display_dict=displayDict,
        )
        metricb.summary_metrics = []
        bundle_list.append(metricb)

    # Add a sky saturation check
    displayDict = {}
    displayDict["group"] = "Basic Stats"
    displayDict["subgroup"] = "Saturation"
    sql = ""
    metric = metrics.SkySaturationMetric()
    summary = metrics.SumMetric()
    slicer = slicers.UniSlicer()
    bundle_list.append(
        metric_bundles.MetricBundle(metric, slicer, sql, summary_metrics=summary, display_dict=displayDict)
    )

    benchmarkArea = 18000
    benchmarkNvisits = 825
    minNvisits = 750
    displayDict = {"group": "SRD", "subgroup": "FO metrics", "order": 0}

    # Configure the count metric which is what is used for f0 slicer.
    metric = metrics.CountExplimMetric(metric_name="fO")
    plotDict = {
        "xlabel": "Number of Visits",
        "asky": benchmarkArea,
        "n_visits": minNvisits,
        "x_min": 0,
        "x_max": 1500,
    }
    summaryMetrics = [
        metrics.FOArea(
            nside=nside,
            norm=False,
            metric_name="fOArea",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FOArea(
            nside=nside,
            norm=True,
            metric_name="fOArea/benchmark",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FONv(
            nside=nside,
            norm=False,
            metric_name="fONv",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FONv(
            nside=nside,
            norm=True,
            metric_name="fONv/benchmark",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FOArea(
            nside=nside,
            norm=False,
            metric_name=f"fOArea_{minNvisits}",
            asky=benchmarkArea,
            n_visit=minNvisits,
        ),
    ]
    caption = "The FO metric evaluates the overall efficiency of observing. "
    caption += (
        "foNv: out of %.2f sq degrees, the area receives at least X and a median of Y visits "
        "(out of %d, if compared to benchmark). " % (benchmarkArea, benchmarkNvisits)
    )
    caption += (
        "fOArea: this many sq deg (out of %.2f sq deg if compared "
        "to benchmark) receives at least %d visits. " % (benchmarkArea, benchmarkNvisits)
    )
    displayDict["caption"] = caption
    slicer = slicers.HealpixSlicer(nside=nside)
    bundle = metric_bundles.MetricBundle(
        metric,
        slicer,
        "",
        plot_dict=plotDict,
        display_dict=displayDict,
        summary_metrics=summaryMetrics,
        plot_funcs=[plots.FOPlot()],
    )
    bundle_list.append(bundle)

    # check that we have coverage for the first year
    displayDict = {}
    displayDict["group"] = "Year 1"
    displayDict["subgroup"] = "Coverage"
    nside_foot = 32
    slicer = slicers.HealpixSlicer(nside=nside_foot, badval=0)
    sky = EuclidOverlapFootprint(nside=nside_foot, smc_radius=4, lmc_radius=6)
    footprints_hp_array, labels = sky.return_maps()
    for filtername in filternames:
        sql = "filter='%s' and night < 365" % filtername
        metric = metrics.CountMetric(col="night", metric_name="N year 1")
        summary_stat = metrics.FootprintFractionMetric(
            footprint=footprints_hp_array[filtername],
            n_min=3,
        )
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            display_dict=displayDict,
            summary_metrics=summary_stat,
            plot_funcs=subsetPlots,
            plot_dict={"color_max": 10},
        )
        bundle_list.append(bundle)

    # Some ToO stats
    displayDict = {"group": "ToO", "order": 1}
    slicer = spatial_slicer(
        nside=nside,
        lat_col=colmap["dec"],
        lon_col=colmap["ra"],
        lat_lon_deg=colmap["raDecDeg"],
    )
    for filtername in filternames:
        sql = "filter='%s' and scheduler_note like 'ToO%%'" % filtername
        metric = metrics.CountMetric(col=colmap["mjd"], metric_name="N ToO")
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            display_dict=displayDict,
            summary_metrics=extended_stats,
            plot_funcs=subsetPlots,
            plot_dict={},
        )
        bundle_list.append(bundle)

    too_sqls = ["scheduler_note like 'ToO, %" + "t%i'" % hour for hour in [0, 1, 2, 4, 24, 48]] + [
        "scheduler_note like 'ToO, %'"
    ]
    slicer = slicers.UniSlicer()
    for sql in too_sqls:
        metric = metrics.CountMetric(col="night")
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            display_dict=displayDict,
        )
        bundle_list.append(bundle)

        metric = metrics.CountUniqueMetric(col="night")
        bundle = metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            display_dict=displayDict,
        )
        bundle_list.append(bundle)

    for b in bundle_list:
        b.set_run_name(run_name)

    bd = metric_bundles.make_bundles_dict_from_list(bundle_list)

    # Add hourglass plots.
    hrDict = hourglassPlots(colmap=colmap, runName=run_name, nyears=nyears, extraSql=sql_constraint)
    bd.update(hrDict)

    # Add basic slew stats.
    try:
        slewDict = slewBasics(colmap=colmap, run_name=run_name)
        bd.update(slewDict)
    except KeyError as e:
        warnings.warn("Could not add slew stats: missing required key %s from colmap" % (e))

    return bd



# Transient metric with input ascii SED.
# Builds upon and uses a significant amount of common code from the
# transientAsciiMetric by: fbb@nyu.edu, svalenti@lcogt.net
# Apart from the inherent differences between metrics, the code has changed in
# that many of the variable names have been changed for clarity, at the expense
# of brevity.
#
# Contact for this code:
# christian.setzer@fysik.su.se

__all__ = ("TransientAsciiSEDMetric",)

import os
from copy import deepcopy

import numpy as np

try:
    from sncosmo import Model, TimeSeriesSource, read_griddata_ascii
except ImportError:
    pass
from astropy.cosmology import Planck15 as cosmo  # noqa N813

from rubin_sim.maf.metrics import BaseMetric
from rubin_sim.maf.utils import m52snr


class TransientAsciiSEDMetric(BaseMetric):
    """
    Based on the transientMetric and transientAsciiMetric, uses an ascii
    input file and provides option to write out the light curves. Calculates
    what fraction of the transients would be detected. A spatial slicer is the
    preferred choice. The input SED should have the 3-column format
    (phase, wave, flux), and be scaled to a distance of 10pc from the observer.

    Parameters
    -----------
    ascii_file : `str`
        The ascii file containing the inputs for the SED. The file must
        contain three columns - ['phase', 'wave', 'flux'] -
        of phase/epoch (in days), wavelength (Angstroms), and
        flux (ergs/s/Angstrom).
    metric_name : `str`, optional
        Name of the metric, can be overwritten by user or child metric.
    survey_duration : `float`, optional
        Length of survey (years).
        Default 10 or maximum of timespan of observations.
    survey_start : `float`, optional
        MJD for the survey start date.
        Default None (uses the time of the first observation at each pointing).
    detect_snr : `dict`, optional
        An observation will be counted toward the discovery criteria if the
        light curve SNR is higher than detect_snr (specified per bandpass).
        Values must be provided for each filter which should be considered
        in the lightcurve.
        Default is {'u': 5, 'g': 5, 'r': 5, 'i': 5, 'z': 5, 'y': 5}
    z : `float`, optional
        Cosmological redshift at which to consider observations of the
        tranisent SED.
    num_pre_time : `int`, optional
        Number of observations (in any filter(s)) to demand before pre_time,
        before saying a transient has been detected.
        Default 0.
    pre_time : `float`, optional
        The time by which num_pre_time detections are required (in days).
        Default 5.0.
    num_filters : `int`, optional
        Number of filters that need to be observed for an object to be
        counted as detected. Default 1. (if num_per_lightcurve is 0, then
        this will be reset to 0).
    filter_time : `float`, optional
        The time within which observations in at least num_filters are
        required (in days). Default None (no time constraint).
    num_per_lightcurve : `int`, optional
        Number of sections of the light curve that must be sampled above
        the detect_snr theshold for the light curve to be counted.
        For example, num_per_lightcurve = 2 means a light curve is only
        considered detected if there is at least 1 observation in the first
        half of the LC, and at least one in the second half of the LC.
        num_per_lightcurve = 4 means each quarter of the light curve must
        be detected to count. Default 1.
    num_phases_to_run : `int`, optional
        Sets the number of phases that should be checked.
        One can imagine pathological cadences where many objects pass the
        detection criteria, but would not if the observations were offset
        by a phase-shift. Default 1.
    output_data : `bool`, optional
        If True, metric returns full lightcurve at each point. Note that
        this will potentially create a very large metric output data file.
        If False, metric returns the number of transients detected.
    """

    def __init__(
        self,
        ascii_file,
        metric_name="TransientAsciiSEDMetric",
        mjd_col="expMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        survey_duration=10.0,
        survey_start=None,
        detect_snr={"u": 5, "g": 5, "r": 5, "i": 5, "z": 5, "y": 5},
        z=0.075,
        num_pre_time=2,
        pre_time=25.0,
        num_filters=2,
        filter_time=None,
        num_per_lightcurve=1,
        num_phases_to_run=5,
        output_data=False,
        **kwargs,
    ):
        # Set all initial attributes of the metric.
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.output_data = output_data
        self.z = z

        # If it is specified that the lightcurves should be returned rather
        # than the summary statistics, this changes the output of the metric,
        # and thus needs an alternate definition of the output type and units
        # when calling the init of the base class.
        if self.output_data:
            super_dict = {"units": "", "metricDtype": "object"}
        else:
            super_dict = {"units": "Number Detected"}

        super(TransientAsciiSEDMetric, self).__init__(
            col=[self.mjd_col, self.m5_col, self.filter_col],
            metric_name=metric_name,
            **super_dict,
            **kwargs,
        )
        # Continue setting the initial attributes.
        self.survey_duration = survey_duration
        self.survey_start = survey_start
        self.detect_snr = detect_snr
        self.num_pre_time = num_pre_time
        self.pre_time = pre_time
        self.num_filters = num_filters
        self.filter_time = filter_time
        self.num_per_lightcurve = num_per_lightcurve
        if self.num_per_lightcurve == 0:
            self.num_filters = 0
        self.num_phases_to_run = num_phases_to_run
        # Read ascii lightcurve template here.
        # It doesn't change per slice_point.
        self.read_sed(ascii_file)

    def read_sed(self, ascii_file):
        """
        Reads in an ascii file detailing the time evolution of an SED. Must be
        in the following format, 3 columns: phase, wavelength, flux.

        This will also set the source model and redshift the model according to
        initialization parameters.

        Parameters
        -----------
        ascii_file : `str`
            string containing the path to the ascii file containing the
            SED evolution.

        """
        # Sanity to check to make sure the provided string indeed points to a
        # file.
        if not os.path.isfile(ascii_file):
            raise IOError("Could not find SED ascii file %s" % (ascii_file))
        # Read in the SED text file with sncosmo for use with it's Model API.
        self.sed_phase, self.sed_wave, self.sed_flux = read_griddata_ascii(ascii_file)
        # Make the source model and set the transient Model to be used for
        # light curve generation and observation.
        self.make_model()
        # Given the redshifted SED set the transient duration.
        self.transient_duration = self.redshifted_model.maxtime() - self.redshifted_model.mintime()

    def make_model(self):
        """
        Wrapper function to take the phase, wave, and flux information from the
        provided ascii file and create an sncosmo Model object,
        and consistently redshift that model given initialization Parameters.
        This sets the transient model in rest frame, and transient model in
        observer frame,
        i.e., it is cosmologically redshifted.
        """
        # Set the source model with sncosmo API.
        source = TimeSeriesSource(self.sed_phase, self.sed_wave, self.sed_flux)
        # Create transient model from sncosmo API
        # Use deepcopy to make ensure full class is saved as attribute of new
        # class.
        self.transient_model = deepcopy(Model(source=source))
        # With the Model set, apply the cosmological redshift specified at
        # initialization.
        self.set_redshift()

    def set_redshift(self):
        """
        Function which takes the input desired observation redshift and
        cosmologically redshifts the source SED. This sets the redshifted_model
        which is what is needed for making observations.
        """

        z = self.z
        # Deepcopy the model to local variable to get proper behavior of model
        # object methods.
        redshifted_model = deepcopy(self.transient_model)
        # Compute the luminosity distance using Planck15 cosmological
        # parameters, by which to decrease the amplitude of the SED
        lumdist = cosmo.luminosity_distance(z).value * 1e6  # in pc
        # SED is assumed to be at 10pc so scale accordingly.
        amp = pow(np.divide(10.0, lumdist), 2)
        # Set the redshift of the SED, this stretches the wavelength
        # distribution.
        redshifted_model.set(z=z)
        # Separately set the amplitude, this rescales the flux at each
        # wavelength.
        redshifted_model.set(amplitude=amp)
        self.redshifted_model = redshifted_model

    def make_lightcurve(self, time, filters):
        """
        Compute light curve magnitudes from the source Model for the specified
        light curve phases, and filters.

        Parameters
        ----------
        time : `np.ndarray`, (N,)
            The times of the observations.
        filters : `list` [`str`]
            The filters of the observations. ['u','g','r',...] format.

        Returns
        -------
        light_curve_mags : `np.ndarray`, (N,)
             The magnitudes of the object at the times and in the filters of
             the observations.
        """
        # Use the redshifted source model.
        redshifted_model = deepcopy(self.redshifted_model)
        # initialize the light curve magnitudes array.
        light_curve_mags = np.zeros(time.size, dtype=float)

        # Iterate over the filters that are observed.
        for flt in list(set(filters)):
            # Find the phases of obesrvations in the current filter.
            flt_times = list(time[np.where(filters == flt)[0]])
            # Initialize the band magnitude list.
            filter_mag = []
            # Get all band magnitudes. Currently must be in loop, method fails
            # in the scipy interpolation if attempting to vectorize according
            # to sncosmo documentation.
            for obs_time in flt_times:
                filter_mag.append(redshifted_model.bandmag("lsst" + flt, "ab", obs_time))
            # Set light_curve_mags for array indices corresponding to
            # observations of the current filter.
            light_curve_mags[np.where(filters == flt)[0]] = np.array(filter_mag)
            self.light_curve_mags = light_curve_mags

    def evaluate_all_detection_criteria(self, data_slice):
        """
        Wrapper function to setup loop for each transient light curve and
        evaluate all detection criteria.

        Parameters
        -----------
        data_slice : `np.array`
            Numpy structured array containing the data related to the visits
            provided by the slicer.

        Returns
        --------
        transient_detected : `np.array`, (`bool`,)
            Array containing `bool` tracking variable whether transient is
            detected by passing all criteria.
        num_detected : `int`
            Scalar value of the number of transients that were detected in
            total between all phase shifts considered.

        """
        # Track whether each individual light curve was detected.
        # Start with the assumption that it is True, and if it fails
        # criteria then becomes False.
        self.transient_detected = np.ones(len(np.unique(self.transient_id)), dtype=bool)

        # Loop through each lightcurve and check if it meets requirements.
        for i, (start_ind, end_ind) in enumerate(zip(self.transient_start_index, self.transient_end_index)):
            t_id = i
            # If there were no observations at all for this lightcurve:
            if start_ind == end_ind:
                self.transient_detected[t_id] = False
                continue

            self.observation_epoch_above_thresh = self.observation_epoch[start_ind:end_ind][
                np.where(self.obs_above_snr_threshold[start_ind:end_ind])
            ]

            self.evaluate_pre_time_detection_criteria(t_id)
            # Check if previous condition passed.
            # If not, move to next transient.
            if not self.transient_detected[t_id]:
                continue

            self.evaluate_phase_section_detection_criteria(t_id)
            # Check if previous condition passed.
            # If not, move to next transient.
            if not self.transient_detected[t_id]:
                continue

            self.evaluate_number_filters_detection_criteria(data_slice, start_ind, end_ind, t_id)
            # Check if previous condition passed.
            # If not, move to next transient.
            if not self.transient_detected[t_id]:
                continue

            self.evaluate_filter_in_time_detection_criteria(t_id)
            # Check if previous condition passed.
            # If not, move to next transient.
            # Note: this last if block is technically unnecessary but if
            # further criteria are added then the if block should be copied
            # afterwards.
            if not self.transient_detected[t_id]:
                continue
            # Finished with current set of conditions

        # Find the unique number of light curves that passed the required
        # number of conditions
        self.num_detected += len(np.where(self.transient_detected == True)[0])

    def evaluate_pre_time_detection_criteria(self, t_id):
        """
        Function evaluate if the specfied number of observations of the current
        transient take place before the user speficied light curve phase by
        which these must be achieved.

        Parameters
        -----------
        t_id : `int`
            The transient id of the currently evaluated transient.
        """
        # If we did not get enough detections before pre_time, set
        # transient_detected to False.
        indices_pre_time = np.where(self.observation_epoch_above_thresh < self.pre_time)[0]
        if len(indices_pre_time) < self.num_pre_time:
            self.transient_detected[t_id] = False

    def evaluate_phase_section_detection_criteria(self, t_id):
        """
        Function to evaluate if the specified number of equal length sections
        of the current transient are detected given the user criteria.

        Parameters
        -----------
        t_id : `int`
            The transient id of the currently evaluated transient.
        """
        # If we did not get detections over enough sections of the
        # lightcurve, set tranisent_detected to False.
        detected_light_curve_sections = np.unique(
            np.floor(self.observation_epoch_above_thresh / self.transient_duration * self.num_per_lightcurve)
        )
        if len(detected_light_curve_sections) < self.num_per_lightcurve:
            self.transient_detected[t_id] = False

    def evaluate_number_filters_detection_criteria(self, data_slice, start_ind, end_ind, t_id):
        """
        Function to evaluate if the current transient passes the required
        number of detections in different filters.

        Parameters
        -----------
        data_slice : `np.array`, (N,)
            Numpy structured array containing the data related to the visits
            provided by the slicer.
        start_ind : `int`
            Starting index for observations of the specific transient being
            evaluated.
        end_ind : `int`
            Ending index for observations of the specific transient being
            evaluated.
        t_id : `int`
            The transient id of the currently evaluated transient.
        """
        # If we did not get detections in enough filters, set transient
        # detected to False.
        self.detected_filters = data_slice[start_ind:end_ind][
            np.where(self.obs_above_snr_threshold[start_ind:end_ind])
        ][self.filter_col]
        if len(np.unique(self.detected_filters)) < self.num_filters:
            self.transient_detected[t_id] = False

    def evaluate_filter_in_time_detection_criteria(self, t_id):
        """
        Function to evaluate whether the required detections in different
        filters take place within the specified time span for them to count
        towards detection of the transient.

        Parameters
        -----------
        t_id : `int`
            The transient id of the currently evaluted transient.
        """
        # If we did not get detections in enough filters within required
        # time, set transient_detected to False.
        if (self.filter_time is not None) and (self.num_filters > 1):
            final_filter_detection_ind = np.searchsorted(
                self.observation_epoch_above_thresh,
                self.observation_epoch_above_thresh + self.filter_time,
                "right",
            )
            final_filter_detection_ind = np.where(
                final_filter_detection_ind < len(self.observation_epoch_above_thresh) - 1,
                final_filter_detection_ind,
                len(self.observation_epoch_above_thresh) - 1,
            )
            is_detected = False
            for i, filt_end_ind in enumerate(final_filter_detection_ind):
                if len(np.unique(self.detected_filters[i:filt_end_ind])) >= self.num_filters:
                    is_detected = True
                    break
            if not is_detected:
                self.transient_detected[t_id] = False

    def setup_phase_shift_dependent_variables(self, time_shift, data_slice):
        """
        Wrapper function to initialize variables that will change for each
        phase shift that is considered.

        Parameters
        -----------
        time_shift : `float`
            The offset given the currently considered phase shift by which
            to cyclically shift the SED evolution.
        data_slice : `np.array`, (N,)
            Numpy structured array containing the data related to the visits
            provided by the slicer.

        Returns
        ----------
        max_num_transients : `int`
            Updated number of the total simulated transients.
        observation_epoch : `np.array`, (N,)
            Array of transient light curve phases of observations of
            transients within this phase shift cycle.
        transient_id : `np.array`, (N,)
            Array of all the transient ids within this phase shift cycle,
            regardless of whether it is observed. dtype int.
        transient_id_start : `int`
            Updated starting id for next phase shift loop.
        transient_start_index : `np.array`, (N,)
            Array of the indicies for each transient that are the start of
            their observations in the observation array. dtype int.
        transient_end_index: `np.array`, (N,)
            Array of the indicies for each transient that are the end of
            their observations in the observation array. dtype int.
        """
        # Update the maximum possible transients that could have been
        # observed during survey_duration.
        self.max_num_transients += np.ceil(self.survey_duration / (self.transient_duration / 365.25))
        # Calculate the observation epoch for each transient lightcurve.
        self.observation_epoch = (
            data_slice[self.mjd_col] - self.survey_start + time_shift
        ) % self.transient_duration
        # Identify the observations which belong to each distinct transient.
        self.transient_id = (
            np.floor((data_slice[self.mjd_col] - self.survey_start) / self.transient_duration)
            + self.transient_id_start
        )
        # Set the starting id number for the next phase shift
        self.transient_id_start = self.transient_id.max()
        # Find the set of uniquely observed transients
        unique_transient_id = np.unique(self.transient_id)
        # Find the starting index for each transient_id.
        self.transient_start_index = np.searchsorted(self.transient_id, unique_transient_id, side="left")
        # Find the ending index of each transient_id.
        self.transient_end_index = np.searchsorted(self.transient_id, unique_transient_id, side="right")

    def setup_run_metric_variables(self, data_slice):
        """
        Wrapper function to handle basic initialization of variables used
        to run this metric.

        Parameters
        -----------
        data_slice : `np.array`, (N,)
            Numpy structured array containing the data related to the visits
            provided by the slicer.

        Returns
        ---------
        data_slice : `np.array`, (N,)
            Now sorted in time.
        survey_duration : `float`
            Defaults to the maximum between the chosen slicer and the user
            specified duration given to the metric.
        survey_start : `float`
            Defaults to user specified, or metric default, however if it is
            not defined sets to the earliest time in the given slicer.
        """
        # Sort the entire data_slice in order of time.
        data_slice.sort(order=self.mjd_col)

        # Check that survey_duration is not larger than the time of
        # observations we obtained. If it is, then the max_num_transients will
        # not be accurate.
        data_slice_time_span = (data_slice[self.mjd_col].max() - data_slice[self.mjd_col].min()) / 365.25
        # Take the maximum time delta, either specified or from the slicer, to
        # be the survey duration.
        self.survey_duration = np.max([data_slice_time_span, self.survey_duration])

        # Set the survey start based on the slicer unless otherwise specified.
        if self.survey_start is None:
            self.survey_start = data_slice[self.mjd_col].min()
        return data_slice

    def initialize_phase_loop_variables(self, data_slice):
        """
        Wrapper function to initialize variables needed for checking all
        transietnts and phase shifts for detected transients.

        Parameters
        -----------
        data_slice : `np.array`, (N,)
            Numpy structured array containing the data related to the visits
            provided by the slicer.

        Returns
        ---------
        time_phase_shifts : `np.array`, (N,)
            The phase offsets over which to iterate detections given the
            specfied number of phases to run.
        num_detected : `int`
            Initialized variable for the number detected, set to zero.
        max_num_transients : `int`
            Initialized variable for the total transients that are simulated
            counting the multiplicity due to phase shifts.
        transient_id_start : `int`
            The starting id for simulated transients that are observed. This
            accounts for if the requested length of the data_slice and the
            number of simulated transient observations mismatch the number
            of transients that fit in the specified survey duration given
            the user specified survey start.
        """
        # Depending on the number of phase shifts to apply and check the
        # detectability, compute the necessary time shifts corresponding to
        # phase division of the lightcurves.
        self.time_phase_shifts = (
            np.arange(self.num_phases_to_run) * self.transient_duration / float(self.num_phases_to_run)
        )
        # Total number of transient which have reached detection thresholds.
        self.num_detected = 0
        # Total number of transients which could possibly be detected,
        # given survey duration and transient duration.
        self.max_num_transients = 0
        # Set this, in case survey_start was set to be much earlier than this
        # data (so we start counting at 0).
        self.transient_id_start = -1 * np.floor(
            (data_slice[self.mjd_col].min() - self.survey_start) / self.transient_duration
        )

    def evaluate_snr_thresholds(self, data_slice):
        """
        Take the given data_slice and the set SNR thresholds for observations
        to be considered in further detections and compute which observations
        pass.

        Parameters
        -----------
        data_slice : `np.array`, (N,)
            Numpy structured array containing the data related to the visits
            provided by the slicer.

        Returns
        --------
        obs_above_SNR_threshold: `np.array`, (N,)
            `bool` array corresponding to all observations and whether or
            not, given their filter specified SNR threshold, they pass this
            thresholding cut.
        """
        # Initilize array for observations below or above SNR threshold
        self.obs_above_snr_threshold = np.zeros(len(self.light_curve_SNRs), dtype=bool)
        # Identify which detections rise above the required SNR threshold
        # in each filter.
        for filt in np.unique(data_slice[self.filter_col]):
            # Find the indices for observations in current filter.
            filter_match = np.where(data_slice[self.filter_col] == filt)[0]
            # Find the subset of the above indices which are above SNR
            # threshold condition, otherwise set threshold bool to False.
            self.obs_above_snr_threshold[filter_match] = np.where(
                self.light_curve_SNRs[filter_match] >= self.detect_snr[filt],
                True,
                False,
            )

    def run(self, data_slice, slice_point=None):
        """
        Calculate the detectability of a transient with the specified SED.

        If self.output_data is True, then returns the full lightcurve for each
        object instead of the total number of transients that are detected.

        Parameters
        ----------
        data_slice : `np.array`, (N,)
            Numpy structured array containing the data related to the visits
            provided by the slicer.
        slice_point : `dict`, optional
            Dictionary containing information about the slice_point currently
            active in the slicer.

        Returns
        -------
        result : `float` or `dict`
            The fraction of transients that could be detected.
            (if output_data is False) Otherwise, a dictionary
            with arrays of 'transient_id', 'lcMag', 'detected', 'expMJD',
            'SNR', 'filter', 'epoch'
        """

        data_slice = self.setup_run_metric_variables(data_slice)
        self.initialize_phase_loop_variables(data_slice)

        # Consider each different 'phase shift' separately.
        # We then just have a series of lightcurves, taking place back-to-back.
        for time_shift in self.time_phase_shifts:
            self.setup_phase_shift_dependent_variables(time_shift, data_slice)

            # Generate the actual light curve magnitudes and SNR
            self.make_lightcurve(self.observation_epoch, data_slice[self.filter_col])
            self.light_curve_SNRs = m52snr(self.light_curve_mags, data_slice[self.m5_col])

            # Check observations above the defined threshold for detection.
            self.evaluate_snr_thresholds(data_slice)
            # With useable observations computed,
            # evaluate all detection criteria
            self.evaluate_all_detection_criteria(data_slice)

        if self.output_data:
            # Output all the light curves, regardless of detection threshold,
            # but indicate which were 'detected'.
            # Only returns for one phase shift, not all.
            return {
                "transient_id": self.transient_id,
                "expMJD": data_slice[self.mjd_col],
                "epoch": self.observation_epoch,
                "filter": data_slice[self.filter_col],
                "lcMag": self.light_curve_mags,
                "SNR": self.light_curve_SNRs,
                "detected": self.transient_detected,
            }
        else:
            return float(self.num_detected)



__all__ = ("SatelliteAvoidBasisFunction",)

import healpy as hp
import numpy as np
import rubin_scheduler.scheduler.basis_functions as bf


class SatelliteAvoidBasisFunction(bf.BaseBasisFunction):
    """Uses satellite position information from the Conditions object
    and then avoids streaks.

    Parameters
    ----------
    forecast_time : `float`
        The time ahead to forecast satellite streaks (minutes).
    smooth_fwhm : `float`
        The smoothing full width half max to use (degrees)
    """

    def __init__(self, nside=32, forecast_time=90.0, smooth_fwhm=3.5):
        super().__init__(nside=nside)
        self.forecast_time = forecast_time / 60.0 / 24  # To days
        self.smooth_fwhm = np.radians(smooth_fwhm)

    def _calc_value(self, conditions, indx=None):
        result = 0
        # find the indices that are relevant
        indx_min = np.min(np.searchsorted(conditions.satellite_mjds, conditions.mjd))
        indx_max = np.max(np.searchsorted(conditions.satellite_mjds, conditions.mjd + self.forecast_time))

        if indx_max > indx_min:
            result = np.sum(conditions.satellite_maps[indx_min:indx_max], axis=0)
            result = hp.smoothing(result, fwhm=self.smooth_fwhm)
            result = hp.ud_grade(result, self.nside)
            result[np.where(result < 0)] = 0
            # Make it negative, so positive weights will result
            # in avoiding satellites
            result *= -1

        return result



__all__ = ("OneDSlicer",)

import warnings
from functools import wraps

import numpy as np

from rubin_sim.maf.plots.oned_plotters import OneDBinnedData
from rubin_sim.maf.stackers import ColInfo
from rubin_sim.maf.utils import optimal_bins

from .base_slicer import BaseSlicer


class OneDSlicer(BaseSlicer):
    """OneD Slicer allows the slicing of data into bins in a single dimension.

    Parameters
    ----------
    slice_col_name : `str`
        The name of the data column to base slicing on (i.e. 'airmass', etc.)
    slice_col_units : `str`, optional
        Set a name for the units of the sliceCol. Used for plotting labels.
    bins : np.ndarray, optional
        The data will be sliced into 'bins': this can be defined as an
        array here. Default None.
    bin_min : `float`, optional
    bin_max : `float`, optional
    bin_size : `float`, optional
        If bins is not defined, then bin_min/bin_max/bin_size can be chosen
        to anchor the slice points.
        Default None.
        Priority goes: bins >> bin_min/bin_max/bin_size >> data values
        (if none of the above are chosen).

    Notes
    -----
    All bins except for the last bin are half-open ([a, b)) while the
    last bin is ([a, b]).
    """

    def __init__(
        self,
        slice_col_name=None,
        slice_col_units=None,
        bins=None,
        bin_min=None,
        bin_max=None,
        bin_size=None,
        verbose=True,
        badval=0,
    ):
        super().__init__(verbose=verbose, badval=badval)
        if slice_col_name is None:
            raise ValueError("slice_col_name cannot be left None - choose a data column to group data by")
        self.slice_col_name = slice_col_name
        self.columns_needed = [slice_col_name]
        # We could try to set up the self.bins here --
        # but it's also possible that
        # these bin_min/max/size values have not been set and
        # should just be set from the data.
        self.bins = bins
        self.bin_min = bin_min
        self.bin_max = bin_max
        self.bin_size = bin_size
        # Forget binmin/max/stepsize if bins was set
        if self.bins is not None:
            if bin_min is not None or bin_max is not None or bin_size is not None:
                warnings.warning(
                    f"Both bins and one of the bin_min/bin_max/bin_size was specified. "
                    f"Using bins ({self.bins} values only."
                )
                self.bin_min = None
                self.bin_max = None
                self.bin_size = None
        # Set the column units
        if slice_col_units is not None:
            self.slice_col_units = slice_col_units
        # Try to determine the column units
        else:
            co = ColInfo()
            self.slice_col_units = co.get_units(self.slice_col_name)
        # Set slicer re-initialize values and default plotFunction
        self.slicer_init = {
            "slice_col_name": self.slice_col_name,
            "slice_col_units": slice_col_units,
            "badval": badval,
            "bin_min": self.bin_min,
            "bin_max": self.bin_max,
            "bin_size": self.bin_size,
            "bins": self.bins,
        }
        self.plot_funcs = [
            OneDBinnedData,
        ]

    def setup_slicer(self, sim_data, maps=None):
        """Set up bins in slicer.

        This happens AFTER sim_data is defined,
        thus typically in the MetricBundleGroup.
        This maps data into the bins;
        it's not a good idea to reuse a OneDSlicer as a result.
        """
        if "bins" in self.slice_points:
            warning_msg = "Warning: this OneDSlicer was already set up once. "
            warning_msg += (
                "Re-setting up a OneDSlicer is unpredictable; at the very least, it "
                "will change the mapping of the simulated data into the data slices. "
                "A safer choice is to use a separate OneDSlicer for each MetricBundle."
            )
            warnings.warn(warning_msg)
        slice_col = sim_data[self.slice_col_name]
        # Set bins from data or specified values,
        # if they were previously defined.
        if self.bins is None:
            # Set bin min/max values (could have been set in __init__)
            if self.bin_min is None:
                self.bin_min = np.nanmin(slice_col)
            if self.bin_max is None:
                self.bin_max = np.nanmax(slice_col)
            # Give warning if bin_min = bin_max,
            # and do something at least slightly reasonable.
            if self.bin_min == self.bin_max:
                warnings.warn(
                    "bin_min = bin_max (maybe your data is single-valued?). "
                    "Increasing bin_max by 1 (or 2*bin_size, if bin_size was set)."
                )
                if self.bin_size is not None:
                    self.bin_max = self.bin_max + 2 * self.bin_size
                else:
                    self.bin_max = self.bin_max + 1
            if self.bin_size is None:
                bins = optimal_bins(slice_col, self.bin_min, self.bin_max)
                nbins = np.round(bins)
                self.bin_size = (self.bin_max - self.bin_min) / float(nbins)
            # Set bins
            self.bins = np.arange(self.bin_min, self.bin_max + self.bin_size / 2.0, self.bin_size, "float")
        # nslice is used to stop iteration and should
        # reflect the usable length of the bins
        self.nslice = len(self.bins) - 1
        # and "shape" refers to the length of the datavalues,
        # and should be one less than # of bins because last
        # binvalue is RH edge only
        self.shape = self.nslice
        # Set slice_point metadata.
        self.slice_points["sid"] = np.arange(self.nslice)
        self.slice_points["bins"] = self.bins
        # Add metadata from map if needed.
        self._run_maps(maps)

        indxs = np.argsort(sim_data[self.slice_col_name])
        data_sorted = sim_data[self.slice_col_name][indxs]

        # Setting up slices such that left_edge <= data < right_edge
        # in each slice.
        left = np.searchsorted(data_sorted, self.bins[0:-1], "left")
        right = np.searchsorted(data_sorted, self.bins[1:], "left")

        self.sim_idxs = [indxs[le:ri] for le, ri in zip(left, right)]

        # Set up _slice_sim_data method for this class.
        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):
            """Slice sim_data on oneD sliceCol, to return relevant
            indexes for slice_point.
            """
            idxs = self.sim_idxs[islice]
            bin_left = self.bins[islice]
            bin_right = self.bins[islice + 1]
            return {
                "idxs": idxs,
                "slice_point": {
                    "sid": islice,
                    "bin_left": bin_left,
                    "bin_right": bin_right,
                },
            }

        setattr(self, "_slice_sim_data", _slice_sim_data)

    def __eq__(self, other_slicer):
        """Evaluate if slicers are equivalent."""
        result = False
        if isinstance(other_slicer, OneDSlicer):
            if self.slice_col_name == other_slicer.slice_col_name:
                # If slicer restored from disk or setup,
                # then 'bins' in slice_points dict.
                # This is preferred method to see if slicers are equal.
                if ("bins" in self.slice_points) & ("bins" in other_slicer.slice_points):
                    result = np.array_equal(other_slicer.slice_points["bins"], self.slice_points["bins"])
                # However, before we 'setup' the slicer with data,
                # the slicers could be equivalent.
                else:
                    if (self.bins is not None) and (other_slicer.bins is not None):
                        result = np.array_equal(self.bins, other_slicer.bins)
                    elif (
                        (self.bin_size is not None)
                        and (self.bin_min is not None) & (self.bin_max is not None)
                        and (other_slicer.bin_size is not None)
                        and (other_slicer.bin_min is not None)
                        and (other_slicer.bin_max is not None)
                    ):
                        if (
                            (self.bin_size == other_slicer.bin_size)
                            and (self.bin_min == other_slicer.bin_min)
                            and (self.bin_max == other_slicer.bin_max)
                        ):
                            result = True
        return result



__all__ = ("NeoDistancePlotter",)

import copy

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Ellipse

from .plot_handler import BasePlotter


class NeoDistancePlotter(BasePlotter):
    """
    Special plotter to calculate and plot the maximum distance an H=22 NEO
    could be observable to, in any particular opsim observation.

    Parameters
    ----------
    step : `float`, optional
        Step size to use for radial bins. Default is 0.01 AU.
    eclip_max, eclip_min : `float`, `float`, optional
        Range of ecliptic latitude values to include when creating the plot.
    """

    def __init__(self, step=0.01, eclip_max=10.0, eclip_min=-10.0):
        self.plot_type = "neoxyPlotter"
        self.object_plotter = True

        self.default_plot_dict = {
            "title": None,
            "xlabel": "X (AU)",
            "ylabel": "Y (AU)",
            "x_min": -1.5,
            "x_max": 1.5,
            "y_min": -0.25,
            "y_max": 2.5,
            "units": "Count",
            "figsize": None,
        }
        self.filter2color = {
            "u": "purple",
            "g": "blue",
            "r": "green",
            "i": "cyan",
            "z": "orange",
            "y": "red",
        }
        self.filter_col_name = "filter"
        self.step = step
        self.eclip_max = np.radians(eclip_max)
        self.eclip_min = np.radians(eclip_min)

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """
        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer.
        user_plot_dict: `dict`
            Dictionary of plot parameters set by user
            (overrides default values).
        fig : `matplotlib.figure.Figure`
            Matplotlib figure number to use. Default = None, starts new figure.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
           Figure with the plot.
        """
        in_plane = np.where(
            (metric_value[0]["eclipLat"] >= self.eclip_min) & (metric_value[0]["eclipLat"] <= self.eclip_max)
        )

        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)

        if fig is None:
            fig = plt.figure(plot_dict["figsize"])
        ax = fig.add_subplot(111)

        planet_props = {"Earth": 1.0, "Venus": 0.72, "Mars": 1.52, "Mercury": 0.39}

        planets = []
        for prop in planet_props:
            planets.append(Ellipse((0, 0), planet_props[prop] * 2, planet_props[prop] * 2, fill=False))

        for planet in planets:
            ax.add_artist(planet)

        # Let's make a 2-d histogram in polar coords,
        # then convert and display in cartisian

        r_step = self.step
        rvec = np.arange(0, plot_dict["x_max"] + r_step, r_step)
        theta_step = np.radians(3.5)
        thetavec = np.arange(0, 2 * np.pi + theta_step, theta_step) - np.pi

        # array to hold histogram values
        H = np.zeros((thetavec.size, rvec.size), dtype=float)

        rgrid, thetagrid = np.meshgrid(rvec, thetavec)

        xgrid = rgrid * np.cos(thetagrid)
        ygrid = rgrid * np.sin(thetagrid)

        for dist, x, y in zip(
            metric_value[0]["MaxGeoDist"][in_plane],
            metric_value[0]["NEOHelioX"][in_plane],
            metric_value[0]["NEOHelioY"][in_plane],
        ):
            theta = np.arctan2(y - 1.0, x)
            diff = np.abs(thetavec - theta)
            theta_to_use = thetavec[np.where(diff == diff.min())]
            # This is a slow where-clause, should be possible to speed it up
            # using
            # np.searchsorted+clever slicing or hist2d to build up the map.
            good = np.where((thetagrid == theta_to_use) & (rgrid <= dist))
            H[good] += 1

        # Set the under value to white
        my_cmap = copy.copy(plt.cm.get_cmap("jet"))
        my_cmap.set_under("w")
        blah = ax.pcolormesh(xgrid, ygrid + 1, H, cmap=my_cmap, vmin=0.001, shading="auto")
        cb = plt.colorbar(blah, ax=ax)
        cb.set_label(plot_dict["units"])

        ax.set_xlabel(plot_dict["xlabel"])
        ax.set_ylabel(plot_dict["ylabel"])
        ax.set_title(plot_dict["title"])
        ax.set_ylim([plot_dict["y_min"], plot_dict["y_max"]])
        ax.set_xlim([plot_dict["x_min"], plot_dict["x_max"]])

        ax.plot([0], [1], marker="o", color="b")
        ax.plot([0], [0], marker="o", color="y")

        return fig



__all__ = ["CheckColorSlope", "ColorSlopeMetric", "ColorSlope2NightMetric"]

import numpy as np
from rubin_scheduler.utils import int_binned_stat

from .base_metric import BaseMetric


class CheckColorSlope(object):
    """Check if the data has a color and a slope

    Parameters
    ----------
    color_length : `float`
        The maximum length of time different filters be observed
        to still count as a color (hours). Default 1 hour.
    slope_length : `float`
        The length of time to demand observations in the
        same filter be greater than (hours). Default 3 hours.
    """

    def __init__(
        self, color_length=1.0, slope_length=3.0, filter_col="filter", mjd_col="observationStartMJD"
    ):
        self.color_length = color_length / 24.0
        self.slope_length = slope_length / 24.0

        self.filter_col = filter_col
        self.mjd_col = mjd_col

    def __call__(self, data_slice):
        has_color = False
        has_slope = False

        if np.size(data_slice) < 3:
            return 0
        filters = data_slice[self.filter_col]

        u_filters = np.unique(filters)

        for filtername in u_filters:
            in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
            time_gap = (
                data_slice[self.mjd_col][in_filt].max() - data_slice[self.mjd_col][in_filt][np.newaxis].min()
            )
            if time_gap >= self.slope_length:
                has_slope = True
                break
        for filtername1 in u_filters:
            for filtername2 in u_filters:
                if filtername1 != filtername2:
                    in_filt1 = np.where(filters == filtername1)[0]
                    in_filt2 = np.where(filters == filtername2)[0]
                    time_gaps = (
                        data_slice[self.mjd_col][in_filt1] - data_slice[self.mjd_col][in_filt2][np.newaxis].T
                    )
                    time_gaps = time_gaps[np.where(time_gaps > 0)]
                    if time_gaps.size > 0:
                        if np.min(time_gaps[np.where(time_gaps > 0)]) <= self.color_length:
                            has_color = True
                            break
        if has_color & has_slope:
            return 1
        else:
            return 0


class ColorSlopeMetric(BaseMetric):
    """How many times do we get a color and slope in a night

    A proxy metric for seeing how many times
    there would be the possibility of identifying and
    classifying a transient.

    Parameters
    ----------
    mag : `dict`
        Dictionary with filternames as keys and minimum depth m5
        magnitudes as values. If None, defaults to mag 20 in ugrizy.
    color_length : `float`
        The maximum length of time different filters be observed
        to still count as a color (hours). Default 1 hour.
    slope_length : `float`
        The length of time to demand observations in the
        same filter be greater than (hours). Default 3 hours.
    """

    def __init__(
        self,
        mag=None,
        night_col="night",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        color_length=1.0,
        slope_length=3.0,
        time_col="observationStartMJD",
        units="#",
        metric_name="ColorSlope",
        **kwargs,
    ):
        cols = [filter_col, night_col, m5_col, time_col]

        if mag is None:
            mag = {"u": 20, "g": 20, "r": 20, "i": 20, "z": 20, "y": 20}

        self.night_col = night_col
        self.filter_col = filter_col
        self.m5_col = m5_col
        self.mag = mag
        self.time_col = time_col

        super().__init__(col=cols, units=units, metric_name=metric_name, **kwargs)

        self.sequence_checker = CheckColorSlope(color_length=color_length, slope_length=slope_length)

    def run(self, data_slice, slice_point=None):
        result = 0
        deep_enough = np.zeros(data_slice.size, dtype=bool)
        for filtername in np.unique(data_slice[self.filter_col]):
            in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
            indx = np.where(data_slice[self.m5_col][in_filt] > self.mag[filtername])[0]
            deep_enough[in_filt[indx]] = True
        data = data_slice[deep_enough]
        if data.size > 0:
            _night, result = int_binned_stat(data[self.night_col], data, statistic=self.sequence_checker)

        return np.sum(result)


class ColorSlope2NightMetric(ColorSlopeMetric):
    """Like ColorSlopeMetric, but span over 2 nights

    Parameters
    ----------
    mag : `dict`
        Dictionary with filternames as keys and minimum depth m5
        magnitudes as values. If None, defaults to mag 20 in ugrizy.
    color_length : `float`
        The maximum length of time different filters be observed
        to still count as a color (hours). Default 1 hour.
    slope_length : `float`
        The length of time to demand observations in the
        same filter be greater than (hours). Default 15 hours.
    """

    def __init__(
        self,
        mag=None,
        night_col="night",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        color_length=1.0,
        slope_length=15.0,
        time_col="observationStartMJD",
        units="#",
        metric_name="ColorSlope2Night",
        **kwargs,
    ):
        super().__init__(
            mag=mag,
            night_col=night_col,
            filter_col=filter_col,
            m5_col=m5_col,
            color_length=color_length,
            slope_length=slope_length,
            time_col=time_col,
            units=units,
            metric_name=metric_name,
            **kwargs,
        )

    def run(self, data_slice, slice_point=None):
        result = 0
        deep_enough = np.zeros(data_slice.size, dtype=bool)
        for filtername in np.unique(data_slice[self.filter_col]):
            in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
            indx = np.where(data_slice[self.m5_col][in_filt] > self.mag[filtername])[0]
            deep_enough[in_filt[indx]] = True
        data = data_slice[deep_enough]
        if data.size > 0:
            # Send in nights as pairs, (0,1) (2,3), (4,5), etc
            night_id = np.floor(data[self.night_col] / 2).astype(int)
            _night, result1 = int_binned_stat(night_id, data, statistic=self.sequence_checker)

            # Now to do pairs (1,2), (3,4)
            night_id = np.ceil(data[self.night_col] / 2).astype(int)
            _night, result2 = int_binned_stat(night_id, data, statistic=self.sequence_checker)

            result = np.sum(result1) + np.sum(result2)

        return result



from .hg_plotters import *
from .hourglass_plotters import *
from .mo_plotters import *
from .nd_plotters import *
from .neo_distance_plotter import *
from .night_pointing_plotter import *
from .oned_plotters import *
from .perceptual_rainbow import *
from .plot_handler import *
from .skyproj_plotters import *
from .spatial_plotters import *
from .special_plotters import *
from .two_d_plotters import *
from .xyplotter import *



import unittest

import numpy as np
from rubin_scheduler.utils import SysEngVals

from rubin_sim.phot_utils import (
    predicted_zeropoint,
    predicted_zeropoint_e2v,
    predicted_zeropoint_hardware,
    predicted_zeropoint_hardware_e2v,
    predicted_zeropoint_hardware_itl,
    predicted_zeropoint_itl,
)


class PredictedZeropointsTst(unittest.TestCase):
    def test_predicted_zeropoints(self):
        bands = ["u", "g", "r", "i", "z", "y"]
        sev = SysEngVals()
        for b in bands:
            zp = predicted_zeropoint(band=b, airmass=1.0, exptime=1)
            self.assertAlmostEqual(zp, sev.zp_t[b], delta=0.005)
            zp_hardware = predicted_zeropoint_hardware(b, exptime=1)
            self.assertTrue(zp < zp_hardware)
            # Check the vendors
            zp_v = predicted_zeropoint_itl(band=b, airmass=1.0, exptime=1)
            self.assertAlmostEqual(zp, zp_v, delta=0.1)
            zp_v = predicted_zeropoint_e2v(band=b, airmass=1.0, exptime=1)
            self.assertAlmostEqual(zp, zp_v, delta=0.1)
            zp_v = predicted_zeropoint_hardware_itl(band=b, exptime=1)
            self.assertAlmostEqual(zp_hardware, zp_v, delta=0.1)
            zp_v = predicted_zeropoint_hardware_e2v(band=b, exptime=1)
            self.assertAlmostEqual(zp_hardware, zp_v, delta=0.1)
            # Check some of the scaling
            zp_test = predicted_zeropoint(band=b, airmass=1.5, exptime=1)
            self.assertTrue(zp > zp_test)
            zp_test = predicted_zeropoint(band=b, airmass=1.0, exptime=30)
            self.assertAlmostEqual(zp, zp_test - 2.5 * np.log10(30), places=7)

            funcs = [predicted_zeropoint, predicted_zeropoint_itl, predicted_zeropoint_e2v]
            for zpfunc in funcs:
                zp = []
                for x in np.arange(1.0, 2.5, 0.1):
                    for exptime in np.arange(1.0, 130, 30):
                        zp.append(zpfunc(b, x, exptime))
                zp = np.array(zp)
                self.assertTrue(zp.max() - zp.min() < 6)
                self.assertTrue(zp.max() < 35)
                self.assertTrue(zp.min() > 25)


if __name__ == "__main__":
    unittest.main()



import os
import sqlite3
import unittest
from tempfile import TemporaryDirectory

import pandas as pd

import rubin_sim.maf.db as db
import rubin_sim.maf.metric_bundles as metric_bundles
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers


class TestScalarHourglassPlots(unittest.TestCase):
    def setUp(self):
        self._data_dir_itself = TemporaryDirectory()
        opsdb_fname = create_sample_visit_db(self._data_dir_itself.name)
        self._out_dir_itself = TemporaryDirectory()
        self.out_dir = self._out_dir_itself.name
        self.results_db = db.ResultsDb(out_dir=self.out_dir)
        self.ops_db = db.OpsimDatabase(opsdb_fname)
        self.slicer = slicers.TimeIntervalSlicer(interval_seconds=3600)
        self.metric = metrics.MedianMetric("seeingFwhmEff")

    @unittest.skip("This is now throwing astropy.utils.iers.iers.IERSRangeError, but only sometimes?")
    def test_month_plot(self):
        sql_constraint = ""
        metric_bundle = metric_bundles.MetricBundle(self.metric, self.slicer, sql_constraint)
        bundle_dict = {1: metric_bundle}
        bundle_group = metric_bundles.MetricBundleGroup(
            bundle_dict,
            self.ops_db,
            out_dir=self.out_dir,
            results_db=self.results_db,
        )
        _ = bundle_group.run_all()
        plot_func = plots.YearHourglassPlot(2023)
        _ = metric_bundle.plot(plotFunc=plot_func)


class TestTimeUseHourglassPlots(unittest.TestCase):
    def setUp(self):
        self._data_dir_itself = TemporaryDirectory()
        opsdb_fname = create_sample_visit_db(self._data_dir_itself.name)

        self._out_dir_itself = TemporaryDirectory()
        self.out_dir = self._out_dir_itself.name
        self.results_db = db.ResultsDb(out_dir=self.out_dir)
        self.ops_db = db.OpsimDatabase(opsdb_fname)
        self.slicer = slicers.BlockIntervalSlicer()
        self.metric = metrics.UseMetric()

    @unittest.skip("XXX-seems like this was erroneously passing before?")
    def test_year_plot(self):
        sql_constraint = ""
        metric_bundle = metric_bundles.MetricBundle(self.metric, self.slicer, sql_constraint)
        bundle_dict = {1: metric_bundle}
        bundle_group = metric_bundles.MetricBundleGroup(
            bundle_dict,
            self.ops_db,
            out_dir=self.out_dir,
            results_db=self.results_db,
        )
        _ = bundle_group.run_all()
        plot_func = plots.YearHourglassUsePlot(2023)
        _ = metric_bundle.plot(plotFunc=plot_func)


def create_sample_visit_db(dir_name):
    fname = os.path.join(dir_name, "opsim.sqlite3")

    sample_visits_dict = {
        "observationStartMJD": {
            75831: 59974.1403819843,
            88243: 59991.13314155574,
            84877: 59986.16509162588,
            73474: 59971.038614223085,
            91024: 59994.3667217889,
            94567: 60001.38094520713,
            88867: 59992.06039742178,
            91397: 59995.17212771153,
            88282: 59991.151247547146,
            82396: 59983.0460827712,
        },
        "visitTime": {
            75831: 34.0,
            88243: 34.0,
            84877: 34.0,
            73474: 34.0,
            91024: 34.0,
            94567: 34.0,
            88867: 34.0,
            91397: 34.0,
            88282: 34.0,
            82396: 34.0,
        },
        "seeingFwhmEff": {
            75831: 0.9580660562409485,
            88243: 1.2816139136535794,
            84877: 0.8711626732163685,
            73474: 1.138078694389459,
            91024: 0.6851416920629385,
            94567: 0.9526705179882091,
            88867: 1.0749841095670385,
            91397: 0.9092889658087193,
            88282: 1.3069907982221605,
            82396: 0.8034796448061345,
        },
        "filter": {
            75831: "y",
            88243: "g",
            84877: "z",
            73474: "r",
            91024: "r",
            94567: "y",
            88867: "r",
            91397: "r",
            88282: "g",
            82396: "r",
        },
        "note": {
            75831: "blob, zy, b",
            88243: "blob, ug, b",
            84877: "blob, iz, b",
            73474: "greedy",
            91024: "greedy",
            94567: "greedy",
            88867: "blob, ur, b",
            91397: "blob, ri, a",
            88282: "blob, gr, a",
            82396: "DD:ECDFS",
        },
    }

    visits = pd.DataFrame(sample_visits_dict)
    with sqlite3.connect(fname) as con:
        visits.to_sql("observations", con=con)

    return fname


run_tests_now = __name__ == "__main__"
if run_tests_now:
    unittest.main()



__all__ = ("QSONumberCountsMetric",)

import os

import healpy as hp
import numpy as np
from rubin_scheduler.data import get_data_dir
from scipy import interpolate

from .base_metric import BaseMetric
from .exgal_m5 import ExgalM5


class QSONumberCountsMetric(BaseMetric):
    """Calculate the number of quasars expected with SNR>=5
    according to the Shen et al. (2020) QLF - model A in the redshift
    range zmin < z < zmax.

    The 5 sigma depths are obtained using the ExgalM5 metric.
    Only quasars fainter than the saturation magnitude are counted.

    By default, zmin is 0.3 and zmax is the minimum between 6.7 and the
    redshift at which the Lyman break matches the effective wavelength
    of the band. For bands izy, zmax is 6.7. This default choice is to
    match Table 10.2 for i-band quasar counts in the LSST Science book.
    """

    def __init__(
        self,
        lsst_filter,
        m5_col="fiveSigmaDepth",
        units="mag",
        extinction_cut=1.0,
        filter_col="filter",
        metric_name="QSONumberCountsMetric",
        qlf_module="Shen20",
        qlf_model="A",
        sed_model="Richards06",
        zmin=0.3,
        zmax=None,
        **kwargs,
    ):
        # Declare the effective wavelengths.
        self.effwavelen = {
            "u": 367.0,
            "g": 482.5,
            "r": 622.2,
            "i": 754.5,
            "z": 869.1,
            "y": 971.0,
        }

        # Dust Extinction limit.
        # Regions with larger extinction and dropped from the counting.
        self.extinction_cut = extinction_cut

        # Save the filter information.
        self.filter_col = filter_col
        self.lsst_filter = lsst_filter

        # Save zmin and zmax, or set zmax to the default value.
        # The default zmax is the lower number between 6.7 and the
        # redshift at which the Lyman break (91.2nm) hits the
        # effective wavelength of the filter.
        # Note that this means that for i, z and y,
        # the default value for zmax is 6.7
        self.zmin = zmin
        if zmax is None:
            zmax = np.min([6.7, self.effwavelen[self.lsst_filter] / 91.2 - 1.0])
        self.zmax = zmax

        # This calculation uses the ExgalM5 metric. So declare that here.
        self.exgal_m5 = ExgalM5(m5_col=m5_col, units=units)

        # Save the input parameters that relate to the QLF model.
        self.qlf_module = qlf_module
        self.qlf_model = qlf_model
        self.sed_model = sed_model

        # Read the long tables, which the number of quasars expected
        # for a given band, qlf_module and qlf_model in a range of
        # redshifts and magnitudes.
        table_name = "Long_Table.LSST{0}.{1}.{2}.{3}.txt".format(
            self.lsst_filter, self.qlf_module, self.qlf_model, self.sed_model
        )
        data_dir = os.path.join(get_data_dir(), "maf", "quasarNumberCounts")
        filename = os.path.join(data_dir, table_name)
        with open(filename, "r") as f:
            mags = np.array([float(x) for x in f.readline().split()])
            zs = np.array([float(x) for x in f.readline().split()])
        mz_data = np.loadtxt(filename, skiprows=2)

        # Make the long table cumulative.
        c_mz_data = np.zeros((mz_data.shape[0] + 1, mz_data.shape[1] + 1))
        c_mz_data[1:, 1:] = mz_data
        c_mz_data = np.cumsum(c_mz_data, axis=0)
        c_mz_data = np.cumsum(c_mz_data, axis=1)

        # Create a 2D interpolation object for the long table.
        # self.nqso_cumulative = interpolate.interp2d(zs[:-1], mags[:-1],
        # #c_mz_data[:-1, :-1], kind="cubic")
        self.nqso_cumulative_aux = interpolate.RectBivariateSpline(
            zs[:-1], mags[:-1], c_mz_data[:-1, :-1].T, kx=3, ky=3
        )

        self.nqso_cumulative = lambda z_new, m_new: self.nqso_cumulative_aux(z_new, m_new).T[0]

        super().__init__(
            col=[m5_col, filter_col, "saturation_mag"],
            metric_name=metric_name,
            maps=self.exgal_m5.maps,
            units=units,
            **kwargs,
        )

    def run(self, data_slice, slice_point=None):
        # exclude areas with high extinction
        if slice_point["ebv"] > self.extinction_cut:
            return self.badval

        # For the data_slice, get the 5 sigma limiting magnitude.
        d_s = data_slice[data_slice[self.filter_col] == self.lsst_filter]
        mlim5 = self.exgal_m5.run(d_s, slice_point)

        # Get the slicer pixel area.
        nside = slice_point["nside"]
        pix_area = hp.nside2pixarea(nside, degrees=True)

        # tranform that limiting magnitude into an expected number of quasars.
        # If there is more than one, take the faintest.
        m_bright = np.max(d_s["saturation_mag"])
        n11 = self.nqso_cumulative(self.zmin, m_bright)
        n12 = self.nqso_cumulative(self.zmin, mlim5)
        n21 = self.nqso_cumulative(self.zmax, m_bright)
        n22 = self.nqso_cumulative(self.zmax, mlim5)

        nqso = (n22 - n21 - n12 + n11) * pix_area
        return nqso
